# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for Volume Code."""

import time
from unittest import mock

import ddt
import os_brick
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import imageutils

from cinder.common import constants
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder import quota
from cinder import quota_utils
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import utils as tests_utils
from cinder.tests.unit import volume as base
import cinder.volume
from cinder.volume import api as volume_api
from cinder.volume.flows.manager import create_volume as create_volume_manager
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_types
from cinder.volume import volume_utils


QUOTAS = quota.QUOTAS

CONF = cfg.CONF


def create_snapshot(volume_id, size=1, metadata=None, ctxt=None,
                    **kwargs):
    """Create a snapshot object."""
    metadata = metadata or {}
    snap = objects.Snapshot(ctxt or context.get_admin_context())
    snap.volume_size = size
    snap.user_id = kwargs.get('user_id', fake.USER_ID)
    snap.project_id = kwargs.get('project_id', fake.PROJECT_ID)
    snap.volume_id = volume_id
    snap.status = fields.SnapshotStatus.CREATING
    if metadata is not None:
        snap.metadata = metadata
    snap.update(kwargs)

    snap.create()
    return snap


@ddt.ddt
class VolumeMigrationTestCase(base.BaseVolumeTestCase):

    def setUp(self):
        super(VolumeMigrationTestCase, self).setUp()
        self._clear_patch = mock.patch(
            'cinder.volume.volume_utils.clear_volume',
            autospec=True)
        self._clear_patch.start()
        self.expected_status = 'available'
        self._service = tests_utils.create_service(
            self.context,
            values={'host': 'newhost', 'binary': constants.VOLUME_BINARY})

    def tearDown(self):
        super(VolumeMigrationTestCase, self).tearDown()
        self._clear_patch.stop()

    def test_migrate_volume_driver(self):
        """Test volume migration done by driver."""
        # Mock driver and rpc functions
        self.mock_object(self.volume.driver, 'migrate_volume',
                         lambda x, y, z, new_type_id=None: (
                             True, {'user_id': fake.USER_ID}))

        volume = tests_utils.create_volume(ctxt=self.context, size=0,
                                           host=CONF.host,
                                           migration_status='migrating')
        host_obj = {'host': 'newhost', 'capabilities': {}}
        self.volume.migrate_volume(self.context, volume, host_obj, False)

        # check volume properties
        volume = objects.Volume.get_by_id(context.get_admin_context(),
                                          volume.id)
        self.assertEqual('newhost', volume.host)
        self.assertEqual('success', volume.migration_status)

    @mock.patch('cinder.volume.manager.VolumeManager.'
                '_can_use_driver_migration')
    def test_migrate_volume_driver_for_retype(self, mock_can_use):
        """Test volume migration done by driver on a retype."""
        # Mock driver and rpc functions
        mock_driver = self.mock_object(self.volume.driver, 'migrate_volume',
                                       return_value=(True, {}))

        volume = tests_utils.create_volume(self.context, size=0,
                                           host=CONF.host,
                                           migration_status='migrating')
        host_obj = {'host': 'newhost', 'capabilities': {}}
        self.volume.migrate_volume(self.context, volume, host_obj, False,
                                   fake.VOLUME_TYPE2_ID, mock.sentinel.diff)

        mock_can_use.assert_called_once_with(mock.sentinel.diff)
        mock_driver.assert_called_once_with(self.context, volume, host_obj)
        # check volume properties
        volume = objects.Volume.get_by_id(context.get_admin_context(),
                                          volume.id)
        self.assertEqual('newhost', volume.host)
        self.assertEqual('success', volume.migration_status)
        self.assertEqual(fake.VOLUME_TYPE2_ID, volume.volume_type_id)

    @mock.patch('cinder.volume.manager.VolumeManager._migrate_volume_generic')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                '_can_use_driver_migration')
    def test_migrate_volume_driver_for_retype_generic(self, mock_can_use,
                                                      mock_generic):
        """Test generic volume migration on a retype after driver can't."""
        # Mock driver and rpc functions
        mock_driver = self.mock_object(self.volume.driver, 'migrate_volume',
                                       return_value=(False, None))

        volume = tests_utils.create_volume(self.context, size=0,
                                           host=CONF.host,
                                           migration_status='migrating')
        host_obj = {'host': 'newhost', 'capabilities': {}}
        self.volume.migrate_volume(self.context, volume, host_obj, False,
                                   fake.VOLUME_TYPE2_ID, mock.sentinel.diff)

        mock_can_use.assert_called_once_with(mock.sentinel.diff)
        mock_driver.assert_called_once_with(self.context, volume, host_obj)
        mock_generic.assert_called_once_with(self.context, volume, host_obj,
                                             fake.VOLUME_TYPE2_ID)

    def test_migrate_volume_driver_cross_az(self):
        """Test volume migration done by driver."""
        # Mock driver and rpc functions
        self.mock_object(self.volume.driver, 'migrate_volume',
                         lambda x, y, z, new_type_id=None: (
                             True, {'user_id': fake.USER_ID}))
        dst_az = 'AZ2'
        db.service_update(self.context, self._service.id,
                          {'availability_zone': dst_az})

        volume = tests_utils.create_volume(self.context, size=0,
                                           host=CONF.host,
                                           migration_status='migrating')
        host_obj = {'host': 'newhost', 'capabilities': {}}
        self.volume.migrate_volume(self.context, volume, host_obj, False)

        # check volume properties
        volume.refresh()
        self.assertEqual('newhost', volume.host)
        self.assertEqual('success', volume.migration_status)
        self.assertEqual(dst_az, volume.availability_zone)

    def _fake_create_volume(self, ctxt, volume, req_spec, filters,
                            allow_reschedule=True):
        return db.volume_update(ctxt, volume['id'],
                                {'status': self.expected_status})

    def test_migrate_volume_error(self):
        with mock.patch.object(self.volume.driver, 'migrate_volume') as \
                mock_migrate,\
                mock.patch.object(self.volume.driver, 'create_export') as \
                mock_create_export:

            # Exception case at self.driver.migrate_volume and create_export
            mock_migrate.side_effect = processutils.ProcessExecutionError
            mock_create_export.side_effect = processutils.ProcessExecutionError
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(processutils.ProcessExecutionError,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              False)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              volume.id)
            self.assertEqual('error', volume.migration_status)
            self.assertEqual('available', volume.status)

    @mock.patch('cinder.compute.API')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                'migrate_volume_completion')
    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_migrate_volume_generic(self, volume_get,
                                    migrate_volume_completion,
                                    nova_api):
        def Volume(original=objects.Volume, **kwargs):
            return original(**kwargs)

        fake_db_new_volume = {'status': 'available', 'id': fake.VOLUME_ID}
        fake_new_volume = fake_volume.fake_db_volume(**fake_db_new_volume)
        new_volume_obj = fake_volume.fake_volume_obj(self.context,
                                                     **fake_new_volume)
        host_obj = {'host': 'newhost', 'capabilities': {}}
        volume_get.return_value = fake_new_volume
        update_server_volume = nova_api.return_value.update_server_volume
        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host)

        volume_mock = self.mock_object(objects, 'Volume', side_effect=Volume)
        with mock.patch.object(self.volume, '_copy_volume_data') as \
                mock_copy_volume:
            self.volume._migrate_volume_generic(self.context, volume,
                                                host_obj, None)

            # Temporary created volume must not use quota
            self.assertFalse(volume_mock.call_args[1]['use_quota'])
            mock_copy_volume.assert_called_with(self.context, volume,
                                                new_volume_obj,
                                                remote='dest')
            migrate_volume_completion.assert_called_with(
                self.context, volume, new_volume_obj, error=False)
            self.assertFalse(update_server_volume.called)

    @mock.patch('cinder.compute.API')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                'migrate_volume_completion')
    def test_migrate_volume_generic_cross_az(self, migrate_volume_completion,
                                             nova_api):
        """Test that we set the right AZ in cross AZ migrations."""
        original_create = objects.Volume.create
        dst_az = 'AZ2'
        db.service_update(self.context, self._service.id,
                          {'availability_zone': dst_az})

        def my_create(self, *args, **kwargs):
            self.status = 'available'
            original_create(self, *args, **kwargs)

        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host)

        host_obj = {'host': 'newhost', 'capabilities': {}}
        create_vol = self.patch('cinder.objects.Volume.create',
                                side_effect=my_create, autospec=True)

        with mock.patch.object(self.volume, '_copy_volume_data') as copy_mock:
            self.volume._migrate_volume_generic(self.context, volume, host_obj,
                                                None)
            copy_mock.assert_called_with(self.context, volume, mock.ANY,
                                         remote='dest')
        migrate_volume_completion.assert_called_with(
            self.context, volume, mock.ANY, error=False)

        nova_api.return_value.update_server_volume.assert_not_called()

        self.assertEqual(dst_az,
                         create_vol.call_args[0][0]['availability_zone'])

    @mock.patch('cinder.compute.API')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                'migrate_volume_completion')
    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_migrate_volume_generic_attached_volume(self, volume_get,
                                                    migrate_volume_completion,
                                                    nova_api):
        attached_host = 'some-host'
        fake_volume_id = fake.VOLUME_ID
        fake_db_new_volume = {'status': 'available', 'id': fake_volume_id}
        fake_new_volume = fake_volume.fake_db_volume(**fake_db_new_volume)
        host_obj = {'host': 'newhost', 'capabilities': {}}
        fake_uuid = fakes.get_fake_uuid()
        update_server_volume = nova_api.return_value.update_server_volume
        volume_get.return_value = fake_new_volume
        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host)
        volume = tests_utils.attach_volume(
            self.context, volume, fake_uuid, attached_host, '/dev/vda')
        self.assertIsNotNone(volume.volume_attachment[0].id)
        self.assertEqual(fake_uuid, volume.volume_attachment[0].instance_uuid)
        self.assertEqual('in-use', volume.status)
        self.volume._migrate_volume_generic(self.context, volume,
                                            host_obj, None)
        self.assertFalse(migrate_volume_completion.called)
        update_server_volume.assert_called_with(self.context, fake_uuid,
                                                volume['id'], fake_volume_id)

    @mock.patch('cinder.objects.volume.Volume.save')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_volume')
    @mock.patch('cinder.compute.API')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                'migrate_volume_completion')
    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_migrate_volume_generic_volume_from_snap(self, volume_get,
                                                     migrate_volume_completion,
                                                     nova_api, create_volume,
                                                     save):
        def fake_create_volume(*args, **kwargs):
            context, volume, request_spec, filter_properties = args
            fake_db = mock.Mock()
            task = create_volume_manager.ExtractVolumeSpecTask(fake_db)
            specs = task.execute(context, volume, {})
            self.assertEqual('raw', specs['type'])

        def fake_copy_volume_data_with_chk_param(*args, **kwargs):
            context, src, dest = args
            self.assertEqual(src['snapshot_id'], dest['snapshot_id'])

        fake_db_new_volume = {'status': 'available', 'id': fake.VOLUME_ID}
        fake_new_volume = fake_volume.fake_db_volume(**fake_db_new_volume)
        host_obj = {'host': 'newhost', 'capabilities': {}}
        volume_get.return_value = fake_new_volume

        volume_from_snap = tests_utils.create_volume(self.context, size=1,
                                                     host=CONF.host)
        volume_from_snap['snapshot_id'] = fake.SNAPSHOT_ID
        create_volume.side_effect = fake_create_volume

        with mock.patch.object(self.volume, '_copy_volume_data') as \
                mock_copy_volume:
            mock_copy_volume.side_effect = fake_copy_volume_data_with_chk_param
            self.volume._migrate_volume_generic(self.context, volume_from_snap,
                                                host_obj, None)

    @mock.patch('cinder.objects.volume.Volume.save')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.create_volume')
    @mock.patch('cinder.compute.API')
    @mock.patch('cinder.volume.manager.VolumeManager.'
                'migrate_volume_completion')
    @mock.patch('cinder.db.sqlalchemy.api.volume_get')
    def test_migrate_volume_generic_for_clone(self, volume_get,
                                              migrate_volume_completion,
                                              nova_api, create_volume, save):
        def fake_create_volume(*args, **kwargs):
            context, volume, request_spec, filter_properties = args
            fake_db = mock.Mock()
            task = create_volume_manager.ExtractVolumeSpecTask(fake_db)
            specs = task.execute(context, volume, {})
            self.assertEqual('raw', specs['type'])

        def fake_copy_volume_data_with_chk_param(*args, **kwargs):
            context, src, dest = args
            self.assertEqual(src['source_volid'], dest['source_volid'])

        fake_db_new_volume = {'status': 'available', 'id': fake.VOLUME_ID}
        fake_new_volume = fake_volume.fake_db_volume(**fake_db_new_volume)
        host_obj = {'host': 'newhost', 'capabilities': {}}
        volume_get.return_value = fake_new_volume

        clone = tests_utils.create_volume(self.context, size=1,
                                          host=CONF.host)
        clone['source_volid'] = fake.VOLUME2_ID
        create_volume.side_effect = fake_create_volume

        with mock.patch.object(self.volume, '_copy_volume_data') as \
                mock_copy_volume:
            mock_copy_volume.side_effect = fake_copy_volume_data_with_chk_param
            self.volume._migrate_volume_generic(self.context, clone,
                                                host_obj, None)

    @mock.patch.object(volume_rpcapi.VolumeAPI, 'update_migrated_volume')
    @mock.patch.object(volume_rpcapi.VolumeAPI, 'delete_volume')
    @mock.patch.object(volume_rpcapi.VolumeAPI, 'create_volume')
    def test_migrate_volume_for_volume_generic(self, create_volume,
                                               rpc_delete_volume,
                                               update_migrated_volume):
        fake_volume = tests_utils.create_volume(self.context, size=1,
                                                previous_status='available',
                                                host=CONF.host)

        host_obj = {'host': 'newhost', 'capabilities': {}}
        with mock.patch.object(self.volume.driver, 'migrate_volume') as \
                mock_migrate_volume,\
                mock.patch.object(self.volume, '_copy_volume_data'),\
                mock.patch.object(self.volume.driver, 'delete_volume') as \
                delete_volume:
            create_volume.side_effect = self._fake_create_volume
            self.volume.migrate_volume(self.context, fake_volume, host_obj,
                                       True)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              fake_volume.id)
            self.assertEqual('newhost', volume.host)
            self.assertEqual('success', volume.migration_status)
            self.assertFalse(mock_migrate_volume.called)
            self.assertFalse(delete_volume.called)
            self.assertTrue(rpc_delete_volume.called)
            self.assertTrue(update_migrated_volume.called)

    def test_migrate_volume_generic_copy_error(self):
        with mock.patch.object(self.volume.driver, 'migrate_volume'),\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'create_volume')\
                as mock_create_volume,\
                mock.patch.object(self.volume, '_copy_volume_data') as \
                mock_copy_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'delete_volume'),\
                mock.patch.object(self.volume, 'migrate_volume_completion'),\
                mock.patch.object(self.volume.driver, 'create_export'):

            # Exception case at migrate_volume_generic
            # source_volume['migration_status'] is 'migrating'
            mock_create_volume.side_effect = self._fake_create_volume
            mock_copy_volume.side_effect = processutils.ProcessExecutionError
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(processutils.ProcessExecutionError,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              True)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              volume.id)
            self.assertEqual('error', volume.migration_status)
            self.assertEqual('available', volume.status)

    @mock.patch('cinder.image.image_utils.qemu_img_info')
    def test_migrate_volume_with_glance_metadata(self, mock_qemu_info):
        volume = self._create_volume_from_image(clone_image_volume=True)
        glance_metadata = volume.glance_metadata

        # We imitate the behavior of rpcapi, by serializing and then
        # deserializing the volume object we created earlier.
        serializer = objects.base.CinderObjectSerializer()
        serialized_volume = serializer.serialize_entity(self.context, volume)
        volume = serializer.deserialize_entity(self.context, serialized_volume)

        image_info = imageutils.QemuImgInfo()
        image_info.virtual_size = '1073741824'
        mock_qemu_info.return_value = image_info

        host_obj = {'host': 'newhost', 'capabilities': {}}
        with mock.patch.object(self.volume.driver,
                               'migrate_volume') as mock_migrate_volume:
            mock_migrate_volume.side_effect = (
                lambda x, y, z, new_type_id=None: (
                    True, {'user_id': fake.USER_ID}))
            self.volume.migrate_volume(self.context, volume, host_obj,
                                       False)
        self.assertEqual('newhost', volume.host)
        self.assertEqual('success', volume.migration_status)
        self.assertEqual(glance_metadata, volume.glance_metadata)

    @mock.patch('cinder.db.volume_update')
    def test_update_migrated_volume(self, volume_update):
        fake_host = 'fake_host'
        fake_new_host = 'fake_new_host'
        fake_update = {'_name_id': fake.VOLUME2_NAME_ID,
                       'provider_location': 'updated_location'}
        fake_elevated = context.RequestContext(fake.USER_ID, self.project_id,
                                               is_admin=True)
        volume = tests_utils.create_volume(self.context, size=1,
                                           status='available',
                                           host=fake_host)
        new_volume = tests_utils.create_volume(
            self.context, size=1,
            status='available',
            provider_location='fake_provider_location',
            _name_id=fake.VOLUME_NAME_ID,
            host=fake_new_host)
        new_volume._name_id = fake.VOLUME_NAME_ID
        new_volume.provider_location = 'fake_provider_location'
        fake_update_error = {'_name_id': new_volume._name_id,
                             'provider_location':
                             new_volume.provider_location}
        expected_update = {'_name_id': volume._name_id,
                           'provider_location': volume.provider_location}
        with mock.patch.object(self.volume.driver,
                               'update_migrated_volume') as migrate_update,\
                mock.patch.object(self.context, 'elevated') as elevated:
            migrate_update.return_value = fake_update
            elevated.return_value = fake_elevated
            self.volume.update_migrated_volume(self.context, volume,
                                               new_volume, 'available')
            volume_update.assert_has_calls((
                mock.call(fake_elevated, new_volume.id, expected_update),
                mock.call(fake_elevated, volume.id, fake_update)))

            # Test the case for update_migrated_volume not implemented
            # for the driver.
            migrate_update.reset_mock()
            volume_update.reset_mock()
            # Reset the volume objects to their original value, since they
            # were changed in the last call.
            new_volume._name_id = fake.VOLUME_NAME_ID
            new_volume.provider_location = 'fake_provider_location'
            migrate_update.side_effect = NotImplementedError
            self.volume.update_migrated_volume(self.context, volume,
                                               new_volume, 'available')
            volume_update.assert_has_calls((
                mock.call(fake_elevated, new_volume.id, fake_update),
                mock.call(fake_elevated, volume.id, fake_update_error)))

    def test_migrate_volume_generic_create_volume_error(self):
        self.expected_status = 'error'

        with mock.patch.object(self.volume.driver, 'migrate_volume'), \
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'create_volume') as mock_create_volume, \
                mock.patch.object(self.volume, '_clean_temporary_volume') as \
                clean_temporary_volume:

            # Exception case at the creation of the new temporary volume
            mock_create_volume.side_effect = self._fake_create_volume
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(exception.VolumeMigrationFailed,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              True)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              volume.id)
            self.assertEqual('error', volume['migration_status'])
            self.assertEqual('available', volume['status'])
            self.assertTrue(clean_temporary_volume.called)
        self.expected_status = 'available'

    def test_migrate_volume_generic_timeout_error(self):
        CONF.set_override("migration_create_volume_timeout_secs", 2)

        with mock.patch.object(self.volume.driver, 'migrate_volume'), \
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'create_volume') as mock_create_volume, \
                mock.patch.object(self.volume, '_clean_temporary_volume') as \
                clean_temporary_volume, \
                mock.patch.object(time, 'sleep'):

            # Exception case at the timeout of the volume creation
            self.expected_status = 'creating'
            mock_create_volume.side_effect = self._fake_create_volume
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(exception.VolumeMigrationFailed,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              True)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              volume.id)
            self.assertEqual('error', volume['migration_status'])
            self.assertEqual('available', volume['status'])
            self.assertTrue(clean_temporary_volume.called)
        self.expected_status = 'available'

    def test_migrate_volume_generic_create_export_error(self):
        with mock.patch.object(self.volume.driver, 'migrate_volume'),\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'create_volume')\
                as mock_create_volume,\
                mock.patch.object(self.volume, '_copy_volume_data') as \
                mock_copy_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'delete_volume'),\
                mock.patch.object(self.volume, 'migrate_volume_completion'),\
                mock.patch.object(self.volume.driver, 'create_export') as \
                mock_create_export:

            # Exception case at create_export
            mock_create_volume.side_effect = self._fake_create_volume
            mock_copy_volume.side_effect = processutils.ProcessExecutionError
            mock_create_export.side_effect = processutils.ProcessExecutionError
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(processutils.ProcessExecutionError,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              True)
            volume = objects.Volume.get_by_id(context.get_admin_context(),
                                              volume.id)
            self.assertEqual('error', volume['migration_status'])
            self.assertEqual('available', volume['status'])

    def test_migrate_volume_generic_migrate_volume_completion_error(self):
        def fake_migrate_volume_completion(ctxt, volume, new_volume,
                                           error=False):
            db.volume_update(ctxt, volume['id'],
                             {'migration_status': 'completing'})
            raise processutils.ProcessExecutionError

        with mock.patch.object(self.volume.driver, 'migrate_volume'),\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'create_volume')\
                as mock_create_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI, 'delete_volume'),\
                mock.patch.object(self.volume, 'migrate_volume_completion')\
                as mock_migrate_compl,\
                mock.patch.object(self.volume.driver, 'create_export'), \
                mock.patch.object(self.volume, '_attach_volume') \
                as mock_attach, \
                mock.patch.object(self.volume, '_detach_volume'), \
                mock.patch.object(os_brick.initiator.connector,
                                  'get_connector_properties') \
                as mock_get_connector_properties, \
                mock.patch.object(volume_utils, 'copy_volume') as mock_copy, \
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'get_capabilities') \
                as mock_get_capabilities:

            # Exception case at delete_volume
            # source_volume['migration_status'] is 'completing'
            mock_create_volume.side_effect = self._fake_create_volume
            mock_migrate_compl.side_effect = fake_migrate_volume_completion
            mock_get_connector_properties.return_value = {}
            mock_attach.side_effect = [{'device': {'path': 'bar'}},
                                       {'device': {'path': 'foo'}}]
            mock_get_capabilities.return_value = {'sparse_copy_volume': True}
            volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host)
            host_obj = {'host': 'newhost', 'capabilities': {}}
            self.assertRaises(processutils.ProcessExecutionError,
                              self.volume.migrate_volume,
                              self.context,
                              volume,
                              host_obj,
                              True)
            volume = db.volume_get(context.get_admin_context(), volume['id'])
            self.assertEqual('error', volume['migration_status'])
            self.assertEqual('available', volume['status'])
            mock_copy.assert_called_once_with('foo', 'bar', 0, '1M',
                                              sparse=True)

    def fake_attach_volume(self, ctxt, volume, instance_uuid, host_name,
                           mountpoint, mode):
        tests_utils.attach_volume(ctxt, volume.id,
                                  instance_uuid, host_name,
                                  '/dev/vda')

    def _test_migrate_volume_completion(self, status='available',
                                        instance_uuid=None, attached_host=None,
                                        retyping=False,
                                        previous_status='available'):

        initial_status = retyping and 'retyping' or status
        old_volume = tests_utils.create_volume(self.context, size=0,
                                               host=CONF.host,
                                               status=initial_status,
                                               migration_status='migrating',
                                               previous_status=previous_status)
        attachment = None
        if status == 'in-use':
            old_volume = tests_utils.attach_volume(self.context, old_volume,
                                                   instance_uuid,
                                                   attached_host,
                                                   '/dev/vda')
            self.assertEqual('in-use', old_volume.status)
            attachment = old_volume.volume_attachment[0]
        target_status = 'target:%s' % old_volume.id
        new_host = CONF.host + 'new'
        new_volume = tests_utils.create_volume(self.context, size=0,
                                               host=new_host,
                                               migration_status=target_status)
        with mock.patch.object(self.volume, 'detach_volume') as \
                mock_detach_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'delete_volume') as mock_delete_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'attach_volume') as mock_attach_volume,\
                mock.patch.object(volume_rpcapi.VolumeAPI,
                                  'update_migrated_volume'),\
                mock.patch.object(self.volume.driver, 'attach_volume'):
            mock_attach_volume.side_effect = self.fake_attach_volume
            old_volume_host = old_volume.host
            new_volume_host = new_volume.host
            self.volume.migrate_volume_completion(self.context, old_volume,
                                                  new_volume)
            after_new_volume = objects.Volume.get_by_id(self.context,
                                                        new_volume.id)
            after_old_volume = objects.Volume.get_by_id(self.context,
                                                        old_volume.id)
            if status == 'in-use':
                mock_detach_volume.assert_called_with(self.context,
                                                      old_volume.id,
                                                      attachment['id'])
                attachments = db.volume_attachment_get_all_by_instance_uuid(
                    self.context, instance_uuid)
                mock_attach_volume.assert_called_once_with(
                    self.context,
                    old_volume,
                    attachment['instance_uuid'],
                    attachment['attached_host'],
                    attachment['mountpoint'],
                    attachment.get('attach_mode', 'rw'),
                )
                self.assertIsNotNone(attachments)
                self.assertEqual(attached_host,
                                 attachments[0]['attached_host'])
                self.assertEqual(instance_uuid,
                                 attachments[0]['instance_uuid'])
            else:
                self.assertFalse(mock_detach_volume.called)
            self.assertTrue(mock_delete_volume.called)
            # NOTE(sborkows): the migrate_volume_completion method alters
            # old and new volume objects, so we need to check the equality
            # between the former host value and the actual one.
            self.assertEqual(old_volume_host, after_new_volume.host)
            self.assertEqual(new_volume_host, after_old_volume.host)

    def test_migrate_volume_completion_retype_available(self):
        self._test_migrate_volume_completion('available', retyping=True)

    def test_migrate_volume_completion_retype_in_use(self):
        self._test_migrate_volume_completion(
            'in-use',
            '83c969d5-065e-4c9c-907d-5394bc2e98e2',
            'some-host',
            retyping=True,
            previous_status='in-use')

    def test_migrate_volume_completion_migrate_available(self):
        self._test_migrate_volume_completion()

    def test_migrate_volume_completion_migrate_in_use(self):
        self._test_migrate_volume_completion(
            'in-use',
            '83c969d5-065e-4c9c-907d-5394bc2e98e2',
            'some-host',
            retyping=False,
            previous_status='in-use')

    @ddt.data(False, True)
    def test_api_migrate_volume_completion_from_swap_with_no_migration(
            self, swap_error):
        # This test validates that Cinder properly finishes the swap volume
        # status updates for the case that no migration has occurred
        instance_uuid = '83c969d5-065e-4c9c-907d-5394bc2e98e2'
        attached_host = 'attached-host'
        orig_attached_vol = tests_utils.create_volume(self.context, size=0)
        orig_attached_vol = tests_utils.attach_volume(
            self.context, orig_attached_vol['id'], instance_uuid,
            attached_host, '/dev/vda')
        new_volume = tests_utils.create_volume(self.context, size=0)

        @mock.patch.object(volume_rpcapi.VolumeAPI, 'detach_volume')
        @mock.patch.object(volume_rpcapi.VolumeAPI, 'attach_volume')
        def _run_migration_completion(rpc_attach_volume,
                                      rpc_detach_volume):
            attachment = orig_attached_vol['volume_attachment'][0]
            attachment_id = attachment['id']
            rpc_attach_volume.side_effect = self.fake_attach_volume
            vol_id = volume_api.API().migrate_volume_completion(
                self.context, orig_attached_vol, new_volume, swap_error)
            if swap_error:
                # When swap failed, we don't want to finish attachment
                self.assertFalse(rpc_detach_volume.called)
                self.assertFalse(rpc_attach_volume.called)
            else:
                # When no error, we should be finishing the attachment
                rpc_detach_volume.assert_called_with(self.context,
                                                     orig_attached_vol,
                                                     attachment_id)
                rpc_attach_volume.assert_called_with(
                    self.context, new_volume, attachment['instance_uuid'],
                    attachment['attached_host'], attachment['mountpoint'],
                    'rw')
            self.assertEqual(new_volume['id'], vol_id)

        _run_migration_completion()

    @mock.patch('cinder.tests.unit.fake_notifier.FakeNotifier._notify')
    def test_retype_setup_fail_volume_is_available(self, mock_notify):
        """Verify volume is still available if retype prepare failed."""
        elevated = context.get_admin_context()
        project_id = self.context.project_id

        db.volume_type_create(elevated, {'name': 'old', 'extra_specs': {}})
        old_vol_type = db.volume_type_get_by_name(elevated, 'old')
        db.volume_type_create(elevated, {'name': 'new', 'extra_specs': {}})
        new_vol_type = db.volume_type_get_by_name(elevated, 'new')
        db.quota_create(elevated, project_id, 'volumes_new', 0)

        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host, status='available',
                                           volume_type_id=old_vol_type['id'])

        api = cinder.volume.api.API()
        self.assertRaises(exception.VolumeLimitExceeded, api.retype,
                          self.context, volume, new_vol_type['id'])

        volume = db.volume_get(elevated, volume.id)
        # FIXME: restore when Bug #1803648 is figured out
        # mock_notify.assert_not_called()
        self.assertEqual('available', volume['status'])

    @mock.patch('cinder.tests.unit.fake_notifier.FakeNotifier._notify')
    def _retype_volume_exec(self, driver, mock_notify,
                            snap=False, policy='on-demand',
                            migrate_exc=False, exc=None, diff_equal=False,
                            replica=False, reserve_vol_type_only=False,
                            encryption_changed=False,
                            replica_new=None):
        elevated = context.get_admin_context()
        project_id = self.context.project_id

        if replica:
            rep_status = 'enabled'
            extra_specs = {'replication_enabled': '<is> True'}
        else:
            rep_status = 'disabled'
            extra_specs = {}

        if replica_new is None:
            replica_new = replica
        new_specs = {'replication_enabled': '<is> True'} if replica_new else {}

        db.volume_type_create(elevated, {'name': 'old',
                                         'extra_specs': extra_specs})
        old_vol_type = db.volume_type_get_by_name(elevated, 'old')

        db.volume_type_create(elevated, {'name': 'new',
                                         'extra_specs': new_specs})
        vol_type = db.volume_type_get_by_name(elevated, 'new')
        new_type = objects.VolumeType.get_by_id(elevated,
                                                vol_type['id'])
        db.quota_create(elevated, project_id, 'volumes_new', 10)

        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host, status='retyping',
                                           volume_type_id=old_vol_type['id'],
                                           replication_status=rep_status)
        volume.previous_status = 'available'
        volume.save()
        if snap:
            snapshot = create_snapshot(volume.id, size=volume.size,
                                       user_id=self.user_context.user_id,
                                       project_id=self.user_context.project_id,
                                       volume_type_id=volume.volume_type_id,
                                       ctxt=self.user_context)
        if driver or diff_equal:
            host_obj = {'host': CONF.host, 'capabilities': {}}
        else:
            host_obj = {'host': 'newhost', 'capabilities': {}}

        reservations = quota_utils.get_volume_type_reservation(
            self.context, volume, new_type.id, reserve_vol_type_only)
        old_reservations = quota_utils.get_volume_type_reservation(
            self.context, volume, old_vol_type['id'], reserve_vol_type_only,
            negative=True)

        old_usage = db.quota_usage_get_all_by_project(elevated, project_id)

        with mock.patch.object(self.volume.driver, 'retype') as _retype,\
                mock.patch.object(volume_types, 'volume_types_diff') as _diff,\
                mock.patch.object(self.volume, 'migrate_volume') as _mig,\
                mock.patch.object(db.sqlalchemy.api, 'volume_get') as _vget,\
                mock.patch.object(context.RequestContext, 'elevated') as _ctx,\
                mock.patch.object(objects.VolumeType, 'get_by_id') as _vtget:
            _vget.return_value = volume
            _retype.return_value = driver
            _ctx.return_value = self.context
            _vtget.return_value = new_type
            returned_diff = {
                'encryption': {},
                'qos_specs': {},
                'extra_specs': {},
            }
            if replica != replica_new:
                returned_diff['extra_specs']['replication_enabled'] = (
                    extra_specs.get('replication_enabled'),
                    new_specs.get('replication_enabled'))
            expected_replica_status = 'enabled' if replica_new else 'disabled'

            if encryption_changed:
                returned_diff['encryption'] = 'fake'
            _diff.return_value = (returned_diff, diff_equal)
            if migrate_exc:
                _mig.side_effect = KeyError
            else:
                _mig.return_value = True

            if not exc:
                self.volume.retype(self.context, volume,
                                   vol_type['id'], host_obj,
                                   migration_policy=policy,
                                   reservations=reservations,
                                   old_reservations=old_reservations)
            else:
                self.assertRaises(exc, self.volume.retype,
                                  self.context, volume,
                                  vol_type['id'], host_obj,
                                  migration_policy=policy,
                                  reservations=reservations,
                                  old_reservations=old_reservations)
            if host_obj['host'] != CONF.host:
                _retype.assert_not_called()

        # get volume/quota properties
        volume = objects.Volume.get_by_id(elevated, volume.id)
        try:
            usage = db.quota_usage_get(elevated, project_id, 'volumes_new')
            volumes_in_use = usage.in_use
        except exception.QuotaUsageNotFound:
            volumes_in_use = 0

        # Get new in_use after retype, it should not be changed.
        if reserve_vol_type_only:
            new_usage = db.quota_usage_get_all_by_project(elevated, project_id)
            for resource in ('volumes', 'gigabytes', 'snapshots'):
                empty = {'in_use': 0, 'reserved': 0}
                # Global resource hasn't changed
                self.assertEqual(old_usage.get(resource, empty)['in_use'],
                                 new_usage.get(resource, empty)['in_use'])
                # The new type was empty before
                self.assertEqual(
                    0, old_usage.get(resource + '_new', empty)['in_use'])
                # Old type resources have been moved to the new one
                self.assertEqual(
                    old_usage.get(resource + '_old', empty)['in_use'],
                    new_usage.get(resource + '_new', empty)['in_use'])
                # The old type is empty now
                self.assertEqual(
                    0, new_usage.get(resource + '_old', empty)['in_use'])

        # check properties
        if driver or diff_equal:
            self.assertEqual(vol_type['id'], volume.volume_type_id)
            self.assertEqual('available', volume.status)
            self.assertEqual(CONF.host, volume.host)
            self.assertEqual(1, volumes_in_use)
            self.assert_notify_called(mock_notify,
                                      (['INFO', 'volume.retype'],),
                                      any_order=True)
            if driver:
                _vtget.assert_called_once_with(self.context,
                                               new_type.id)
                _retype.assert_called_once_with(self.context,
                                                mock.ANY,
                                                new_type,
                                                returned_diff,
                                                host_obj)
            # When retyping a volume with snapshots the snapshots should be
            # retyped as well
            if snap:
                snapshot.refresh()
                self.assertEqual(new_type.id, snapshot.volume_type_id)

        elif not exc:
            self.assertEqual(old_vol_type['id'], volume.volume_type_id)
            self.assertEqual('retyping', volume.status)
            self.assertEqual(CONF.host, volume.host)
            self.assertEqual(1, volumes_in_use)
            self.assert_notify_called(mock_notify,
                                      (['INFO', 'volume.retype'],),
                                      any_order=True)
        else:
            self.assertEqual(old_vol_type['id'], volume.volume_type_id)
            self.assertEqual('available', volume.status)
            self.assertEqual(CONF.host, volume.host)
            self.assertEqual(0, volumes_in_use)
        if encryption_changed:
            self.assertTrue(_mig.called)
        self.assertEqual(expected_replica_status, volume.replication_status)

    def test_retype_volume_driver_success(self):
        self._retype_volume_exec(True)

    @ddt.data((False, False), (False, True), (True, False), (True, True))
    @ddt.unpack
    def test_retype_volume_replica(self, replica, replica_new):
        self._retype_volume_exec(True, replica=replica,
                                 replica_new=replica_new)

    def test_retype_volume_migration_bad_policy(self):
        # Test volume retype that requires migration by not allowed
        self._retype_volume_exec(False, policy='never',
                                 exc=exception.VolumeMigrationFailed)

    def test_retype_volume_migration_with_replica(self):
        self._retype_volume_exec(False,
                                 replica=True,
                                 exc=exception.InvalidVolume)

    def test_retype_volume_migration_with_snaps(self):
        self._retype_volume_exec(False, snap=True, exc=exception.InvalidVolume)

    def test_retype_volume_migration_failed(self):
        self._retype_volume_exec(False, migrate_exc=True, exc=KeyError)

    def test_retype_volume_migration_success(self):
        self._retype_volume_exec(False, migrate_exc=False, exc=None)

    def test_retype_volume_migration_equal_types(self):
        self._retype_volume_exec(False, diff_equal=True)

    def test_retype_volume_migration_equal_types_snaps(self):
        self._retype_volume_exec(False, snap=True, diff_equal=True,
                                 reserve_vol_type_only=True)

    def test_retype_volume_with_type_only(self):
        self._retype_volume_exec(True, reserve_vol_type_only=True)

    def test_retype_volume_migration_encryption(self):
        self._retype_volume_exec(False, encryption_changed=True)

    def test_migrate_driver_not_initialized(self):
        volume = tests_utils.create_volume(self.context, size=0,
                                           host=CONF.host)
        host_obj = {'host': 'newhost', 'capabilities': {}}

        self.volume.driver._initialized = False
        self.assertRaises(exception.DriverNotInitialized,
                          self.volume.migrate_volume,
                          self.context, volume, host_obj, True)

        volume = objects.Volume.get_by_id(context.get_admin_context(),
                                          volume.id)
        self.assertEqual('error', volume.migration_status)

        # lets cleanup the mess.
        self.volume.driver._initialized = True
        self.volume.delete_volume(self.context, volume)

    def test_delete_source_volume_in_migration(self):
        """Test deleting a source volume that is in migration."""
        self._test_delete_volume_in_migration('migrating')

    def test_delete_destination_volume_in_migration(self):
        """Test deleting a destination volume that is in migration."""
        self._test_delete_volume_in_migration('target:vol-id')

    def _test_delete_volume_in_migration(self, migration_status):
        """Test deleting a volume that is in migration."""
        volume = tests_utils.create_volume(self.context, host=CONF.host,
                                           migration_status=migration_status)
        self.volume.delete_volume(self.context, volume=volume)

        # The volume is successfully removed during the volume delete
        # and won't exist in the database any more.
        self.assertRaises(exception.VolumeNotFound, volume.refresh)

    def test_retype_volume_not_capable_to_replica(self):
        elevated = context.get_admin_context()
        db.volume_type_create(elevated, {'name': 'old', 'extra_specs': {}})
        old_vol_type = db.volume_type_get_by_name(elevated, 'old')
        new_extra_specs = {'replication_enabled': '<is> True'}
        new_type = db.volume_type_create(elevated, {'name': 'new',
                                         'extra_specs': new_extra_specs})
        new_vol_type = objects.VolumeType.get_by_id(self.context,
                                                    new_type['id'])
        volume = tests_utils.create_volume(self.context, size=1,
                                           host=CONF.host, status='available',
                                           volume_type_id=old_vol_type['id'],
                                           replication_status='not-capable')
        host_obj = {'host': 'newhost', 'capabilities': {}}
        with mock.patch.object(self.volume,
                               'migrate_volume') as migrate_volume,\
                mock.patch.object(objects.VolumeType,
                                  'get_by_id') as vt_get:
            migrate_volume.return_value = True
            vt_get.return_value = new_vol_type
            self.volume.retype(self.context, volume, new_vol_type.id,
                               host_obj, migration_policy='on-demand')
            vt_get.assert_not_called()

    @ddt.data(
        (None, True),
        ({'encryption': {'cipher': ('v1', 'v2')}}, False),
        ({'qos_specs': {'key1': ('v1', 'v2')}}, False),
        ({'encryption': {}, 'qos_specs': {}, 'extra_specs': {}}, True),
        ({'encryption': {}, 'qos_specs': {},
          'extra_specs': {'volume_backend_name': ('ceph1', 'ceph2'),
                          'RESKEY:availability_zones': ('nova', 'nova2')}},
         True),
        ({'encryption': {}, 'qos_specs': {},
          'extra_specs': {'thin_provisioning_support': ('<is> True', None)}},
         False),
    )
    @ddt.unpack
    def test__can_use_driver_migration(self, diff, expected):
        res = self.volume._can_use_driver_migration(diff)
        self.assertEqual(expected, res)



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2014 TrilioData, Inc
# Copyright (c) 2015 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The backups api."""
from http import HTTPStatus

from oslo_log import log as logging
from oslo_utils import strutils
from webob import exc

from cinder.api import api_utils
from cinder.api import common
from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import backups as backup
from cinder.api import validation
from cinder.api.views import backups as backup_views
from cinder import backup as backupAPI
from cinder import exception
from cinder import utils
from cinder import volume as volumeAPI

LOG = logging.getLogger(__name__)


class BackupsController(wsgi.Controller):
    """The Backups API controller for the OpenStack API."""

    _view_builder_class = backup_views.ViewBuilder

    def __init__(self):
        self.backup_api = backupAPI.API()
        self.volume_api = volumeAPI.API()
        super(BackupsController, self).__init__()

    def show(self, req, id):
        """Return data about the given backup."""
        LOG.debug('Show backup with id: %s.', id)
        context = req.environ['cinder.context']

        # Not found exception will be handled at the wsgi level
        backup = self.backup_api.get(context, backup_id=id)
        req.cache_db_backup(backup)

        return self._view_builder.detail(req, backup)

    @wsgi.response(HTTPStatus.ACCEPTED)
    def delete(self, req, id):
        """Delete a backup."""
        context = req.environ['cinder.context']

        LOG.info('Delete backup with id: %s.', id)

        try:
            backup = self.backup_api.get(context, id)
            self.backup_api.delete(context, backup)
        # Not found exception will be handled at the wsgi level
        except exception.InvalidBackup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

    def index(self, req):
        """Returns a summary list of backups."""
        return self._get_backups(req, is_detail=False)

    def detail(self, req):
        """Returns a detailed list of backups."""
        return self._get_backups(req, is_detail=True)

    @staticmethod
    def _get_backup_filter_options():
        """Return volume search options allowed by non-admin."""
        return ('name', 'status', 'volume_id')

    @common.process_general_filtering('backup')
    def _process_backup_filtering(self, context=None, filters=None,
                                  req_version=None):
        api_utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_backup_filter_options())

    def _convert_sort_name(self, req_version, sort_keys):
        """Convert sort key "name" to "display_name". """
        pass

    def _get_backups(self, req, is_detail):
        """Returns a list of backups, transformed through view builder."""
        context = req.environ['cinder.context']
        filters = req.params.copy()
        req_version = req.api_version_request
        marker, limit, offset = common.get_pagination_params(filters)
        sort_keys, sort_dirs = common.get_sort_params(filters)

        show_count = False
        if req_version.matches(
                mv.SUPPORT_COUNT_INFO) and 'with_count' in filters:
            show_count = utils.get_bool_param('with_count', filters)
            filters.pop('with_count')
        self._convert_sort_name(req_version, sort_keys)
        self._process_backup_filtering(context=context, filters=filters,
                                       req_version=req_version)

        if 'name' in filters:
            filters['display_name'] = filters.pop('name')

        backups = self.backup_api.get_all(context, search_opts=filters.copy(),
                                          marker=marker,
                                          limit=limit,
                                          offset=offset,
                                          sort_keys=sort_keys,
                                          sort_dirs=sort_dirs,
                                          )

        total_count = None
        if show_count:
            total_count = self.volume_api.calculate_resource_count(
                context, 'backup', filters)
        req.cache_db_backups(backups.objects)

        if is_detail:
            backups = self._view_builder.detail_list(req, backups.objects,
                                                     total_count)
        else:
            backups = self._view_builder.summary_list(req, backups.objects,
                                                      total_count)
        return backups

    # TODO(frankm): Add some checks here including
    # - whether requested volume_id exists so we can return some errors
    #   immediately
    # - maybe also do validation of swift container name
    @wsgi.response(HTTPStatus.ACCEPTED)
    @validation.schema(backup.create, mv.BASE_VERSION,
                       mv.get_prior_version(mv.BACKUP_METADATA))
    @validation.schema(backup.create_backup_v343, mv.BACKUP_METADATA,
                       mv.get_prior_version(mv.BACKUP_AZ))
    @validation.schema(backup.create_backup_v351, mv.BACKUP_AZ)
    def create(self, req, body):
        """Create a new backup."""
        LOG.debug('Creating new backup %s', body)

        context = req.environ['cinder.context']
        req_version = req.api_version_request

        backup = body['backup']
        container = backup.get('container', None)
        volume_id = backup['volume_id']

        self.validate_name_and_description(backup, check_length=False)
        name = backup.get('name', None)
        description = backup.get('description', None)
        incremental = strutils.bool_from_string(backup.get(
            'incremental', False), strict=True)
        force = strutils.bool_from_string(backup.get(
            'force', False), strict=True)
        snapshot_id = backup.get('snapshot_id', None)
        metadata = backup.get('metadata', None) if req_version.matches(
            mv.BACKUP_METADATA) else None

        if req_version.matches(mv.BACKUP_AZ):
            availability_zone = backup.get('availability_zone', None)
        else:
            availability_zone = None
        az_text = ' in az %s' % availability_zone if availability_zone else ''

        LOG.info("Creating backup of volume %(volume_id)s in container"
                 " %(container)s%(az)s",
                 {'volume_id': volume_id, 'container': container,
                  'az': az_text},
                 context=context)

        try:
            new_backup = self.backup_api.create(context, name, description,
                                                volume_id, container,
                                                incremental, availability_zone,
                                                force, snapshot_id, metadata)
        except (exception.InvalidVolume,
                exception.InvalidSnapshot,
                exception.InvalidVolumeMetadata,
                exception.InvalidVolumeMetadataSize) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        # Other not found exceptions will be handled at the wsgi level
        except exception.ServiceNotFound as error:
            raise exc.HTTPServiceUnavailable(explanation=error.msg)

        retval = self._view_builder.summary(req, dict(new_backup))
        return retval

    @wsgi.response(HTTPStatus.ACCEPTED)
    @validation.schema(backup.restore)
    def restore(self, req, id, body):
        """Restore an existing backup to a volume."""
        LOG.debug('Restoring backup %(backup_id)s (%(body)s)',
                  {'backup_id': id, 'body': body})

        context = req.environ['cinder.context']
        restore = body['restore']
        volume_id = restore.get('volume_id', None)
        name = restore.get('name', None)

        LOG.info("Restoring backup %(backup_id)s to volume %(volume_id)s.",
                 {'backup_id': id, 'volume_id': volume_id},
                 context=context)

        try:
            new_restore = self.backup_api.restore(context,
                                                  backup_id=id,
                                                  volume_id=volume_id,
                                                  name=name)
        # Not found exception will be handled at the wsgi level
        except (exception.InvalidInput,
                exception.InvalidVolume,
                exception.InvalidBackup) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except (exception.VolumeSizeExceedsAvailableQuota,
                exception.VolumeLimitExceeded) as error:
            raise exc.HTTPRequestEntityTooLarge(
                explanation=error.msg, headers={'Retry-After': '0'})

        retval = self._view_builder.restore_summary(
            req, dict(new_restore))
        return retval

    def export_record(self, req, id):
        """Export a backup."""
        LOG.debug('Export record for backup %s.', id)
        context = req.environ['cinder.context']

        try:
            backup_info = self.backup_api.export_record(context, id)
        # Not found exception will be handled at the wsgi level
        except exception.InvalidBackup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        retval = self._view_builder.export_summary(
            req, dict(backup_info))
        LOG.debug('Exported record output: %s.', retval)
        return retval

    @wsgi.response(HTTPStatus.CREATED)
    @validation.schema(backup.import_record)
    def import_record(self, req, body):
        """Import a backup."""
        LOG.debug('Importing record from %s.', body)
        context = req.environ['cinder.context']
        import_data = body['backup-record']
        backup_service = import_data['backup_service']
        backup_url = import_data['backup_url']

        LOG.debug('Importing backup using %(service)s and url %(url)s.',
                  {'service': backup_service, 'url': backup_url})

        try:
            new_backup = self.backup_api.import_record(context,
                                                       backup_service,
                                                       backup_url)
        except exception.InvalidBackup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        # Other Not found exceptions will be handled at the wsgi level
        except exception.ServiceNotFound as error:
            raise exc.HTTPServiceUnavailable(explanation=error.msg)

        retval = self._view_builder.summary(req, dict(new_backup))
        LOG.debug('Import record output: %s.', retval)
        return retval


class Backups(extensions.ExtensionDescriptor):
    """Backups support."""

    name = 'Backups'
    alias = 'backups'
    updated = '2012-12-12T00:00:00+00:00'

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension(
            Backups.alias, BackupsController(),
            collection_actions={'detail': 'GET', 'import_record': 'POST'},
            member_actions={'restore': 'POST', 'export_record': 'GET',
                            'action': 'POST'})
        resources.append(res)
        return resources



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Mock unit tests for the NetApp block storage driver interfaces"""

from collections import abc

from cinder.tests.unit import test
from cinder.volume.drivers.netapp.dataontap import block_cmode
from cinder.volume.drivers.netapp.dataontap import fc_cmode
from cinder.volume.drivers.netapp.dataontap import iscsi_cmode


class NetAppBlockStorageDriverInterfaceTestCase(test.TestCase):

    def setUp(self):
        super(NetAppBlockStorageDriverInterfaceTestCase, self).setUp()

        self.mock_object(block_cmode.NetAppBlockStorageCmodeLibrary,
                         '__init__',
                         return_value=None)
        self.iscsi_cmode_driver = iscsi_cmode.NetAppCmodeISCSIDriver()
        self.fc_cmode_driver = fc_cmode.NetAppCmodeFibreChannelDriver()

    def test_driver_interfaces_match(self):
        """Ensure the NetApp block storage driver interfaces match.

        The two block storage Cinder drivers from NetApp (iSCSI/FC)
        are merely passthrough shim layers atop a common
        block storage library. Bugs have been introduced when a Cinder
        method was exposed via a subset of those driver shims.  This test
        ensures they remain in sync and the library features are uniformly
        available in the four drivers.
        """

        # Get local functions of each driver interface
        iscsi_cmode = self._get_local_functions(self.iscsi_cmode_driver)
        fc_cmode = self._get_local_functions(self.fc_cmode_driver)

        # Ensure NetApp block storage driver shims are identical
        self.assertSetEqual(iscsi_cmode, fc_cmode)

    def _get_local_functions(self, obj):
        """Get function names of an object without superclass functions."""
        return set([key for key, value in type(obj).__dict__.items()
                    if isinstance(value, abc.Callable)])



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re

from eventlet import greenthread
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils.secretutils import md5
from oslo_utils import units
import six

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import interface
import cinder.privsep.fs
from cinder.volume.drivers.nexenta import jsonrpc
from cinder.volume.drivers.nexenta import options
from cinder.volume.drivers.nexenta import utils
from cinder.volume.drivers import nfs

VERSION = '1.3.1'
LOG = logging.getLogger(__name__)


@interface.volumedriver
class NexentaNfsDriver(nfs.NfsDriver):  # pylint: disable=R0921
    """Executes volume driver commands on Nexenta Appliance.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver version.
        1.1.0 - Auto sharing for enclosing folder.
        1.1.1 - Added caching for NexentaStor appliance 'volroot' value.
        1.1.2 - Ignore "folder does not exist" error in delete_volume and
                delete_snapshot method.
        1.1.3 - Redefined volume_backend_name attribute inherited from
                RemoteFsDriver.
        1.2.0 - Added migrate and retype methods.
        1.3.0 - Extend volume method.
        1.3.1 - Cache capacity info and check shared folders on setup.
    """

    driver_prefix = 'nexenta'
    volume_backend_name = 'NexentaNfsDriver'
    VERSION = VERSION
    VOLUME_FILE_NAME = 'volume'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Nexenta_CI"

    def __init__(self, *args, **kwargs):
        super(NexentaNfsDriver, self).__init__(*args, **kwargs)
        if self.configuration:
            self.configuration.append_config_values(
                options.NEXENTA_CONNECTION_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_NFS_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_DATASET_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_RRMGR_OPTS)

        self.nms_cache_volroot = self.configuration.nexenta_nms_cache_volroot
        self.rrmgr_compression = self.configuration.nexenta_rrmgr_compression
        self.rrmgr_tcp_buf_size = self.configuration.nexenta_rrmgr_tcp_buf_size
        self.rrmgr_connections = self.configuration.nexenta_rrmgr_connections
        self.nfs_mount_point_base = self.configuration.nexenta_mount_point_base
        self.volume_compression = (
            self.configuration.nexenta_dataset_compression)
        self.volume_deduplication = self.configuration.nexenta_dataset_dedup
        self.volume_description = (
            self.configuration.nexenta_dataset_description)
        self.sparsed_volumes = self.configuration.nexenta_sparsed_volumes
        self._nms2volroot = {}
        self.share2nms = {}
        self.nfs_versions = {}
        self.shares_with_capacities = {}

    @staticmethod
    def get_driver_options():
        return (
            options.NEXENTA_CONNECTION_OPTS +
            options.NEXENTA_NFS_OPTS +
            options.NEXENTA_DATASET_OPTS +
            options.NEXENTA_RRMGR_OPTS
        )

    @property
    def backend_name(self):
        backend_name = None
        if self.configuration:
            backend_name = self.configuration.safe_get('volume_backend_name')
        if not backend_name:
            backend_name = self.__class__.__name__
        return backend_name

    def do_setup(self, context):
        shares_config = getattr(self.configuration, self.driver_prefix +
                                '_shares_config')
        if shares_config:
            self.configuration.nfs_shares_config = shares_config
        super(NexentaNfsDriver, self).do_setup(context)
        self._load_shares_config(shares_config)
        self._mount_subfolders()

    def check_for_setup_error(self):
        """Verify that the volume for our folder exists.

        :raise: :py:exc:`LookupError`
        """
        if self.share2nms:
            for nfs_share in self.share2nms:
                nms = self.share2nms[nfs_share]
                volume_name, dataset = self._get_share_datasets(nfs_share)
                if not nms.volume.object_exists(volume_name):
                    raise LookupError(_("Volume %s does not exist in Nexenta "
                                        "Store appliance"), volume_name)
                folder = '%s/%s' % (volume_name, dataset)
                if not nms.folder.object_exists(folder):
                    raise LookupError(_("Folder %s does not exist in Nexenta "
                                        "Store appliance"), folder)
                if (folder not in nms.netstorsvc.get_shared_folders(
                        'svc:/network/nfs/server:default', '')):
                    self._share_folder(nms, volume_name, dataset)
                self._get_capacity_info(nfs_share)

    def migrate_volume(self, ctxt, volume, host):
        """Migrate if volume and host are managed by Nexenta appliance.

        :param ctxt: context
        :param volume: a dictionary describing the volume to migrate
        :param host: a dictionary describing the host to migrate to
        """
        LOG.debug('Enter: migrate_volume: id=%(id)s, host=%(host)s',
                  {'id': volume['id'], 'host': host})

        false_ret = (False, None)

        if volume['status'] not in ('available', 'retyping'):
            LOG.warning("Volume status must be 'available' or 'retyping'."
                        " Current volume status: %s", volume['status'])
            return false_ret

        if 'capabilities' not in host:
            LOG.warning("Unsupported host. No capabilities found")
            return false_ret

        capabilities = host['capabilities']
        ns_shares = capabilities['ns_shares']
        dst_parts = capabilities['location_info'].split(':')
        dst_host, dst_volume = dst_parts[1:]

        if (capabilities.get('vendor_name') != 'Nexenta' or
                dst_parts[0] != self.__class__.__name__ or
                capabilities['free_capacity_gb'] < volume['size']):
            return false_ret

        nms = self.share2nms[volume['provider_location']]
        ssh_bindings = nms.appliance.ssh_list_bindings()
        shares = []
        for bind in ssh_bindings:
            for share in ns_shares:
                if (share.startswith(ssh_bindings[bind][3]) and
                        ns_shares[share] >= volume['size']):
                    shares.append(share)
        if len(shares) == 0:
            LOG.warning("Remote NexentaStor appliance at %s should be "
                        "SSH-bound.", share)
            return false_ret
        share = sorted(shares, key=ns_shares.get, reverse=True)[0]
        snapshot = {
            'volume_name': volume['name'],
            'volume_id': volume['id'],
            'name': utils.get_migrate_snapshot_name(volume)
        }
        self.create_snapshot(snapshot)
        location = volume['provider_location']
        src = '%(share)s/%(volume)s@%(snapshot)s' % {
            'share': location.split(':')[1].split('volumes/')[1],
            'volume': volume['name'],
            'snapshot': snapshot['name']
        }
        dst = ':'.join([dst_host, dst_volume.split('/volumes/')[1]])
        try:
            nms.appliance.execute(self._get_zfs_send_recv_cmd(src, dst))
        except utils.NexentaException as exc:
            LOG.warning("Cannot send source snapshot %(src)s to "
                        "destination %(dst)s. Reason: %(exc)s",
                        {'src': src, 'dst': dst, 'exc': exc})
            return false_ret
        finally:
            try:
                self.delete_snapshot(snapshot)
            except utils.NexentaException as exc:
                LOG.warning("Cannot delete temporary source snapshot "
                            "%(src)s on NexentaStor Appliance: %(exc)s",
                            {'src': src, 'exc': exc})
        try:
            self.delete_volume(volume)
        except utils.NexentaException as exc:
            LOG.warning("Cannot delete source volume %(volume)s on "
                        "NexentaStor Appliance: %(exc)s",
                        {'volume': volume['name'], 'exc': exc})

        dst_nms = self._get_nms_for_url(capabilities['nms_url'])
        dst_snapshot = '%s/%s@%s' % (dst_volume.split('volumes/')[1],
                                     volume['name'], snapshot['name'])
        try:
            dst_nms.snapshot.destroy(dst_snapshot, '')
        except utils.NexentaException as exc:
            LOG.warning("Cannot delete temporary destination snapshot "
                        "%(dst)s on NexentaStor Appliance: %(exc)s",
                        {'dst': dst_snapshot, 'exc': exc})
        return True, {'provider_location': share}

    def _get_zfs_send_recv_cmd(self, src, dst):
        """Returns rrmgr command for source and destination."""
        return utils.get_rrmgr_cmd(src, dst,
                                   compression=self.rrmgr_compression,
                                   tcp_buf_size=self.rrmgr_tcp_buf_size,
                                   connections=self.rrmgr_connections)

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info.

        :param volume: volume reference
        :param connector: connector reference
        """
        export = '%s/%s' % (volume['provider_location'], volume['name'])
        data = {'export': export, 'name': 'volume'}
        if volume['provider_location'] in self.shares:
            data['options'] = self.shares[volume['provider_location']]
        return {
            'driver_volume_type': self.driver_volume_type,
            'data': data
        }

    def retype(self, context, volume, new_type, diff, host):
        """Convert the volume to be of the new type.

        :param context: Context
        :param volume: A dictionary describing the volume to migrate
        :param new_type: A dictionary describing the volume type to convert to
        :param diff: A dictionary with the difference between the two types
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug('Retype volume request %(vol)s to be %(type)s '
                  '(host: %(host)s), diff %(diff)s.',
                  {'vol': volume['name'],
                   'type': new_type,
                   'host': host,
                   'diff': diff})

        options = dict(
            compression='compression',
            dedup='dedup',
            description='nms:description'
        )

        retyped = False
        migrated = False
        model_update = None

        src_backend = self.__class__.__name__
        dst_backend = host['capabilities']['location_info'].split(':')[0]
        if src_backend != dst_backend:
            LOG.warning('Cannot retype from %(src_backend)s to '
                        '%(dst_backend)s.',
                        {'src_backend': src_backend,
                         'dst_backend': dst_backend})
            return False

        hosts = (volume['host'], host['host'])
        old, new = hosts
        if old != new:
            migrated, provider_location = self.migrate_volume(
                context, volume, host)

        if not migrated:
            provider_location = volume['provider_location']
            nms = self.share2nms[provider_location]
        else:
            nms_url = host['capabilities']['nms_url']
            nms = self._get_nms_for_url(nms_url)
            model_update = provider_location
            provider_location = provider_location['provider_location']

        share = provider_location.split(':')[1].split('volumes/')[1]
        folder = '%(share)s/%(volume)s' % {
            'share': share,
            'volume': volume['name']
        }

        for opt in options:
            old, new = diff.get('extra_specs').get(opt, (False, False))
            if old != new:
                LOG.debug('Changing %(opt)s from %(old)s to %(new)s.',
                          {'opt': opt, 'old': old, 'new': new})
                try:
                    nms.folder.set_child_prop(
                        folder, options[opt], new)
                    retyped = True
                except utils.NexentaException:
                    LOG.error('Error trying to change %(opt)s'
                              ' from %(old)s to %(new)s',
                              {'opt': opt, 'old': old, 'new': new})
                    return False, None
        return retyped or migrated, model_update

    def _do_create_volume(self, volume):
        nfs_share = volume['provider_location']
        nms = self.share2nms[nfs_share]

        vol, dataset = self._get_share_datasets(nfs_share)
        folder = '%s/%s' % (dataset, volume['name'])
        LOG.debug('Creating folder on Nexenta Store %s', folder)
        nms.folder.create_with_props(
            vol, folder,
            {'compression': self.configuration.nexenta_dataset_compression}
        )

        volume_path = self.remote_path(volume)
        volume_size = volume['size']
        try:
            self._share_folder(nms, vol, folder)

            if getattr(self.configuration,
                       self.driver_prefix + '_sparsed_volumes'):
                self._create_sparsed_file(nms, volume_path, volume_size)
            else:
                folder_path = '%s/%s' % (vol, folder)
                compression = nms.folder.get_child_prop(
                    folder_path, 'compression')
                if compression != 'off':
                    # Disable compression, because otherwise will not use space
                    # on disk.
                    nms.folder.set_child_prop(
                        folder_path, 'compression', 'off')
                try:
                    self._create_regular_file(nms, volume_path, volume_size)
                finally:
                    if compression != 'off':
                        # Backup default compression value if it was changed.
                        nms.folder.set_child_prop(
                            folder_path, 'compression', compression)

            self._set_rw_permissions_for_all(nms, volume_path)

            if self._get_nfs_server_version(nfs_share) < 4:
                sub_share, mnt_path = self._get_subshare_mount_point(nfs_share,
                                                                     volume)
                self._ensure_share_mounted(sub_share, mnt_path)
            self._get_capacity_info(nfs_share)
        except utils.NexentaException:
            try:
                nms.folder.destroy('%s/%s' % (vol, folder))
            except utils.NexentaException:
                LOG.warning("Cannot destroy created folder: "
                            "%(vol)s/%(folder)s",
                            {'vol': vol, 'folder': folder})
            raise

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create new volume from other's snapshot on appliance.

        :param volume: reference of volume to be created
        :param snapshot: reference of source snapshot
        """
        self._ensure_shares_mounted()

        snapshot_vol = self._get_snapshot_volume(snapshot)
        nfs_share = snapshot_vol['provider_location']
        volume['provider_location'] = nfs_share
        nms = self.share2nms[nfs_share]

        vol, dataset = self._get_share_datasets(nfs_share)
        snapshot_name = '%s/%s/%s@%s' % (vol, dataset, snapshot['volume_name'],
                                         snapshot['name'])
        folder = '%s/%s' % (dataset, volume['name'])
        nms.folder.clone(snapshot_name, '%s/%s' % (vol, folder))

        try:
            self._share_folder(nms, vol, folder)
        except utils.NexentaException:
            try:
                nms.folder.destroy('%s/%s' % (vol, folder), '')
            except utils.NexentaException:
                LOG.warning("Cannot destroy cloned folder: "
                            "%(vol)s/%(folder)s",
                            {'vol': vol, 'folder': folder})
            raise

        if self._get_nfs_server_version(nfs_share) < 4:
            sub_share, mnt_path = self._get_subshare_mount_point(nfs_share,
                                                                 volume)
            self._ensure_share_mounted(sub_share, mnt_path)

        if (('size' in volume) and (
                volume['size'] > snapshot['volume_size'])):
            self.extend_volume(volume, volume['size'])

        return {'provider_location': volume['provider_location']}

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume.

        :param volume: new volume reference
        :param src_vref: source volume reference
        """
        LOG.info('Creating clone of volume: %s', src_vref['id'])
        snapshot = {'volume_name': src_vref['name'],
                    'volume_id': src_vref['id'],
                    'volume_size': src_vref['size'],
                    'name': self._get_clone_snapshot_name(volume)}
        # We don't delete this snapshot, because this snapshot will be origin
        # of new volume. This snapshot will be automatically promoted by NMS
        # when user will delete its origin.
        self.create_snapshot(snapshot)
        try:
            return self.create_volume_from_snapshot(volume, snapshot)
        except utils.NexentaException:
            LOG.error('Volume creation failed, deleting created snapshot '
                      '%(volume_name)s@%(name)s', snapshot)
            try:
                self.delete_snapshot(snapshot)
            except (utils.NexentaException, exception.SnapshotIsBusy):
                LOG.warning('Failed to delete zfs snapshot '
                            '%(volume_name)s@%(name)s', snapshot)
            raise

    def delete_volume(self, volume):
        """Deletes a logical volume.

        :param volume: volume reference
        """
        nfs_share = volume.get('provider_location')
        if nfs_share:
            nms = self.share2nms[nfs_share]
            vol, parent_folder = self._get_share_datasets(nfs_share)
            folder = '%s/%s/%s' % (vol, parent_folder, volume['name'])
            mount_path = self.remote_path(volume).strip(
                '/%s' % self.VOLUME_FILE_NAME)
            if mount_path in self._remotefsclient._read_mounts():
                cinder.privsep.fs.umount(mount_path)
            try:
                props = nms.folder.get_child_props(folder, 'origin') or {}
                nms.folder.destroy(folder, '-r')
            except utils.NexentaException as exc:
                if 'does not exist' in exc.args[0]:
                    LOG.info('Folder %s does not exist, it was '
                             'already deleted.', folder)
                    return
                raise
            self._get_capacity_info(nfs_share)
            origin = props.get('origin')
            if origin and self._is_clone_snapshot_name(origin):
                try:
                    nms.snapshot.destroy(origin, '')
                except utils.NexentaException as exc:
                    if 'does not exist' in exc.args[0]:
                        LOG.info('Snapshot %s does not exist, it was '
                                 'already deleted.', origin)
                        return
                    raise

    def extend_volume(self, volume, new_size):
        """Extend an existing volume.

        :param volume: volume reference
        :param new_size: volume new size in GB
        """
        LOG.info('Extending volume: %(id)s New size: %(size)s GB',
                 {'id': volume['id'], 'size': new_size})
        nfs_share = volume['provider_location']
        nms = self.share2nms[nfs_share]
        volume_path = self.remote_path(volume)
        if getattr(self.configuration,
                   self.driver_prefix + '_sparsed_volumes'):
            self._create_sparsed_file(nms, volume_path, new_size)
        else:
            block_size_mb = 1
            block_count = ((new_size - volume['size']) * units.Gi /
                           (block_size_mb * units.Mi))

            nms.appliance.execute(
                'dd if=/dev/zero seek=%(seek)d of=%(path)s'
                ' bs=%(bs)dM count=%(count)d' % {
                    'seek': volume['size'] * units.Gi / block_size_mb,
                    'path': volume_path,
                    'bs': block_size_mb,
                    'count': block_count
                }
            )

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: snapshot reference
        """
        volume = self._get_snapshot_volume(snapshot)
        nfs_share = volume['provider_location']
        nms = self.share2nms[nfs_share]
        vol, dataset = self._get_share_datasets(nfs_share)
        folder = '%s/%s/%s' % (vol, dataset, volume['name'])
        nms.folder.create_snapshot(folder, snapshot['name'], '-r')

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: snapshot reference
        """
        volume = self._get_snapshot_volume(snapshot)
        nfs_share = volume['provider_location']
        nms = self.share2nms[nfs_share]
        vol, dataset = self._get_share_datasets(nfs_share)
        folder = '%s/%s/%s' % (vol, dataset, volume['name'])
        try:
            nms.snapshot.destroy('%s@%s' % (folder, snapshot['name']), '')
        except utils.NexentaException as exc:
            if 'does not exist' in exc.args[0]:
                LOG.info('Snapshot %(folder)s@%(snapshot)s does not '
                         'exist, it was already deleted.',
                         {'folder': folder,
                          'snapshot': snapshot})
                return
            elif 'has dependent clones' in exc.args[0]:
                LOG.info('Snapshot %(folder)s@%(snapshot)s has dependent '
                         'clones, it will be deleted later.',
                         {'folder': folder,
                          'snapshot': snapshot})
                return

    def _create_sparsed_file(self, nms, path, size):
        """Creates file with 0 disk usage.

        :param nms: nms object
        :param path: path to new file
        :param size: size of file
        """
        nms.appliance.execute(
            'truncate --size %(size)dG %(path)s' % {
                'path': path,
                'size': size
            }
        )

    def _create_regular_file(self, nms, path, size):
        """Creates regular file of given size.

        Takes a lot of time for large files.

        :param nms: nms object
        :param path: path to new file
        :param size: size of file
        """
        block_size_mb = 1
        block_count = size * units.Gi / (block_size_mb * units.Mi)

        LOG.info('Creating regular file: %s.'
                 'This may take some time.', path)

        nms.appliance.execute(
            'dd if=/dev/zero of=%(path)s bs=%(bs)dM count=%(count)d' % {
                'path': path,
                'bs': block_size_mb,
                'count': block_count
            }
        )

        LOG.info('Regular file: %s created.', path)

    def _set_rw_permissions_for_all(self, nms, path):
        """Sets 666 permissions for the path.

        :param nms: nms object
        :param path: path to file
        """
        nms.appliance.execute('chmod ugo+rw %s' % path)

    def local_path(self, volume):
        """Get volume path (mounted locally fs path) for given volume.

        :param volume: volume reference
        """
        nfs_share = volume['provider_location']
        return os.path.join(self._get_mount_point_for_share(nfs_share),
                            volume['name'], 'volume')

    def _get_mount_point_for_share(self, nfs_share):
        """Returns path to mount point NFS share.

        :param nfs_share: example 172.18.194.100:/var/nfs
        """
        nfs_share = nfs_share.encode('utf-8')
        return os.path.join(self.configuration.nexenta_mount_point_base,
                            md5(nfs_share, usedforsecurity=False).hexdigest())

    def remote_path(self, volume):
        """Get volume path (mounted remotely fs path) for given volume.

        :param volume: volume reference
        """
        nfs_share = volume['provider_location']
        share = nfs_share.split(':')[1].rstrip('/')
        return '%s/%s/volume' % (share, volume['name'])

    def _share_folder(self, nms, volume, folder):
        """Share NFS folder on NexentaStor Appliance.

        :param nms: nms object
        :param volume: volume name
        :param folder: folder name
        """
        path = '%s/%s' % (volume, folder.lstrip('/'))
        share_opts = {
            'read_write': '*',
            'read_only': '',
            'root': 'nobody',
            'extra_options': 'anon=0',
            'recursive': 'true',
            'anonymous_rw': 'true',
        }
        LOG.debug('Sharing folder %s on Nexenta Store', folder)
        nms.netstorsvc.share_folder('svc:/network/nfs/server:default', path,
                                    share_opts)

    def _load_shares_config(self, share_file):
        self.shares = {}
        self.share2nms = {}

        for share in self._read_config_file(share_file):
            # A configuration line may be either:
            # host:/share_name  http://user:pass@host:[port]/
            # or
            # host:/share_name  http://user:pass@host:[port]/
            #    -o options=123,rw --other
            if not share.strip():
                continue
            if share.startswith('#'):
                continue

            share_info = re.split(r'\s+', share, 2)

            share_address = share_info[0].strip()
            nms_url = share_info[1].strip()
            share_opts = share_info[2].strip() if len(share_info) > 2 else None

            if not re.match(r'.+:/.+', share_address):
                LOG.warning("Share %s ignored due to invalid format. "
                            "Must be of form address:/export.",
                            share_address)
                continue

            self.shares[share_address] = share_opts
            self.share2nms[share_address] = self._get_nms_for_url(nms_url)

        LOG.debug('Shares loaded: %s', self.shares)

    def _get_subshare_mount_point(self, nfs_share, volume):
        mnt_path = '%s/%s' % (
            self._get_mount_point_for_share(nfs_share), volume['name'])
        sub_share = '%s/%s' % (nfs_share, volume['name'])
        return sub_share, mnt_path

    def _ensure_share_mounted(self, nfs_share, mount_path=None):
        """Ensure that NFS share is mounted on the host.

        Unlike the parent method this one accepts mount_path as an optional
        parameter and uses it as a mount point if provided.

        :param nfs_share: NFS share name
        :param mount_path: mount path on the host
        """
        mnt_flags = []
        if self.shares.get(nfs_share) is not None:
            mnt_flags = self.shares[nfs_share].split()
        num_attempts = max(1, self.configuration.nfs_mount_attempts)
        for attempt in range(num_attempts):
            try:
                if mount_path is None:
                    self._remotefsclient.mount(nfs_share, mnt_flags)
                else:
                    if mount_path in self._remotefsclient._read_mounts():
                        LOG.info('Already mounted: %s', mount_path)
                        return

                    fileutils.ensure_tree(mount_path)
                    self._remotefsclient._mount_nfs(nfs_share, mount_path,
                                                    mnt_flags)
                return
            except Exception as e:
                if attempt == (num_attempts - 1):
                    LOG.error('Mount failure for %(share)s after '
                              '%(count)d attempts.',
                              {'share': nfs_share,
                               'count': num_attempts})
                    raise exception.NfsException(six.text_type(e))
                LOG.warning(
                    'Mount attempt %(attempt)d failed: %(error)s. '
                    'Retrying mount ...',
                    {'attempt': attempt, 'error': e})
                greenthread.sleep(1)

    def _mount_subfolders(self):
        ctxt = context.get_admin_context()
        vol_entries = self.db.volume_get_all_by_host(ctxt, self.host)
        for vol in vol_entries:
            nfs_share = vol['provider_location']
            if ((nfs_share in self.shares) and
                    (self._get_nfs_server_version(nfs_share) < 4)):
                sub_share, mnt_path = self._get_subshare_mount_point(
                    nfs_share, vol)
                self._ensure_share_mounted(sub_share, mnt_path)

    def _get_nfs_server_version(self, share):
        if not self.nfs_versions.get(share):
            nms = self.share2nms[share]
            nfs_opts = nms.netsvc.get_confopts(
                'svc:/network/nfs/server:default', 'configure')
            try:
                self.nfs_versions[share] = int(
                    nfs_opts['nfs_server_versmax']['current'])
            except KeyError:
                self.nfs_versions[share] = int(
                    nfs_opts['server_versmax']['current'])
        return self.nfs_versions[share]

    def _get_capacity_info(self, nfs_share):
        """Calculate available space on the NFS share.

        :param nfs_share: example 172.18.194.100:/var/nfs
        """
        nms = self.share2nms[nfs_share]
        ns_volume, ns_folder = self._get_share_datasets(nfs_share)
        folder_props = nms.folder.get_child_props('%s/%s' % (ns_volume,
                                                             ns_folder),
                                                  'used|available')
        free = utils.str2size(folder_props['available'])
        allocated = utils.str2size(folder_props['used'])
        self.shares_with_capacities[nfs_share] = {
            'free': utils.str2gib_size(free),
            'total': utils.str2gib_size(free + allocated)}
        return free + allocated, free, allocated

    def _get_nms_for_url(self, url):
        """Returns initialized nms object for url."""
        auto, scheme, user, password, host, port, path = (
            utils.parse_nms_url(url))
        return jsonrpc.NexentaJSONProxy(scheme, host, port, path, user,
                                        password, auto=auto)

    def _get_snapshot_volume(self, snapshot):
        ctxt = context.get_admin_context()
        return db.volume_get(ctxt, snapshot['volume_id'])

    def _get_volroot(self, nms):
        """Returns volroot property value from NexentaStor appliance."""
        if not self.nms_cache_volroot:
            return nms.server.get_prop('volroot')
        if nms not in self._nms2volroot:
            self._nms2volroot[nms] = nms.server.get_prop('volroot')
        return self._nms2volroot[nms]

    def _get_share_datasets(self, nfs_share):
        nms = self.share2nms[nfs_share]
        volroot = self._get_volroot(nms)
        path = nfs_share.split(':')[1][len(volroot):].strip('/')
        volume_name = path.split('/')[0]
        folder_name = '/'.join(path.split('/')[1:])
        return volume_name, folder_name

    def _get_clone_snapshot_name(self, volume):
        """Return name for snapshot that will be used to clone the volume."""
        return 'cinder-clone-snapshot-%(id)s' % volume

    def _is_clone_snapshot_name(self, snapshot):
        """Check if snapshot is created for cloning."""
        name = snapshot.split('@')[-1]
        return name.startswith('cinder-clone-snapshot-')

    def _update_volume_stats(self):
        """Retrieve stats info for NexentaStor appliance."""
        LOG.debug('Updating volume stats')
        total_space = 0
        free_space = 0
        share = None
        for _share in self._mounted_shares:
            if self.shares_with_capacities[_share]['free'] > free_space:
                free_space = self.shares_with_capacities[_share]['free']
                total_space = self.shares_with_capacities[_share]['total']
                share = _share

        location_info = '%(driver)s:%(share)s' % {
            'driver': self.__class__.__name__,
            'share': share
        }
        nms_url = self.share2nms[share].url
        self._stats = {
            'vendor_name': 'Nexenta',
            'dedup': self.volume_deduplication,
            'compression': self.volume_compression,
            'description': self.volume_description,
            'nms_url': nms_url,
            'ns_shares': self.shares_with_capacities,
            'driver_version': self.VERSION,
            'storage_protocol': 'NFS',
            'total_capacity_gb': total_space,
            'free_capacity_gb': free_space,
            'reserved_percentage': self.configuration.reserved_percentage,
            'QoS_support': False,
            'location_info': location_info,
            'volume_backend_name': self.backend_name,
            'nfs_mount_point_base': self.nfs_mount_point_base
        }



#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
from http import HTTPStatus

from castellan import key_manager
from oslo_config import cfg
import oslo_messaging as messaging
from oslo_utils import strutils
import webob

from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_actions as volume_action
from cinder.api import validation
from cinder import exception
from cinder.i18n import _
from cinder.policies import volume_actions as policy
from cinder import volume
from cinder.volume import volume_utils


CONF = cfg.CONF


class VolumeActionsController(wsgi.Controller):
    def __init__(self, *args, **kwargs):
        super(VolumeActionsController, self).__init__(*args, **kwargs)
        self._key_mgr = None
        self.volume_api = volume.API()

    @property
    def _key_manager(self):
        # Allows for lazy initialization of the key manager
        if self._key_mgr is None:
            self._key_mgr = key_manager.API(CONF)

        return self._key_mgr

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-attach')
    @validation.schema(volume_action.attach)
    def _attach(self, req, id, body):
        """Add attachment metadata."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        # instance UUID is an option now
        instance_uuid = None
        if 'instance_uuid' in body['os-attach']:
            instance_uuid = body['os-attach']['instance_uuid']
        host_name = None
        # Keep API backward compatibility
        if 'host_name' in body['os-attach']:
            host_name = body['os-attach']['host_name']
        mountpoint = body['os-attach']['mountpoint']
        mode = body['os-attach'].get('mode', 'rw')

        try:
            self.volume_api.attach(context, volume,
                                   instance_uuid, host_name, mountpoint, mode)
        except messaging.RemoteError as error:
            if error.exc_type in ['InvalidVolume', 'InvalidUUID',
                                  'InvalidVolumeAttachMode']:
                msg = _("Error attaching volume - %(err_type)s: "
                        "%(err_msg)s") % {
                    'err_type': error.exc_type, 'err_msg': error.value}
                raise webob.exc.HTTPBadRequest(explanation=msg)
            else:
                # There are also few cases where attach call could fail due to
                # db or volume driver errors. These errors shouldn't be exposed
                # to the user and in such cases it should raise 500 error.
                raise

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-detach')
    @validation.schema(volume_action.detach)
    def _detach(self, req, id, body):
        """Clear attachment metadata."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        attachment_id = None
        attachment_id = body['os-detach'].get('attachment_id', None)

        try:
            self.volume_api.detach(context, volume, attachment_id)
        except messaging.RemoteError as error:
            if error.exc_type in ['VolumeAttachmentNotFound', 'InvalidVolume']:
                msg = _("Error detaching volume - %(err_type)s: "
                        "%(err_msg)s") % {
                    'err_type': error.exc_type, 'err_msg': error.value}
                raise webob.exc.HTTPBadRequest(explanation=msg)
            else:
                # There are also few cases where detach call could fail due to
                # db or volume driver errors. These errors shouldn't be exposed
                # to the user and in such cases it should raise 500 error.
                raise

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-reserve')
    def _reserve(self, req, id, body):
        """Mark volume as reserved."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        self.volume_api.reserve_volume(context, volume)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-unreserve')
    def _unreserve(self, req, id, body):
        """Unmark volume as reserved."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        self.volume_api.unreserve_volume(context, volume)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-begin_detaching')
    def _begin_detaching(self, req, id, body):
        """Update volume status to 'detaching'."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        self.volume_api.begin_detaching(context, volume)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-roll_detaching')
    def _roll_detaching(self, req, id, body):
        """Roll back volume status to 'in-use'."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        self.volume_api.roll_detaching(context, volume)

    @wsgi.action('os-initialize_connection')
    @validation.schema(volume_action.initialize_connection)
    def _initialize_connection(self, req, id, body):
        """Initialize volume attachment."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)
        connector = body['os-initialize_connection']['connector']
        try:
            info = self.volume_api.initialize_connection(context,
                                                         volume,
                                                         connector)
        except exception.InvalidInput as err:
            raise webob.exc.HTTPBadRequest(
                explanation=err.msg)
        except exception.VolumeBackendAPIException:
            msg = _("Unable to fetch connection information from backend.")
            raise webob.exc.HTTPInternalServerError(explanation=msg)
        except messaging.RemoteError as error:
            if error.exc_type == 'InvalidInput':
                raise exception.InvalidInput(reason=error.value)
            raise

        return {'connection_info': info}

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-terminate_connection')
    @validation.schema(volume_action.terminate_connection)
    def _terminate_connection(self, req, id, body):
        """Terminate volume attachment."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)
        connector = body['os-terminate_connection']['connector']
        try:
            self.volume_api.terminate_connection(context, volume, connector)
        except exception.VolumeBackendAPIException:
            msg = _("Unable to terminate volume connection from backend.")
            raise webob.exc.HTTPInternalServerError(explanation=msg)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-volume_upload_image')
    @validation.schema(volume_action.volume_upload_image, mv.BASE_VERSION,
                       mv.get_prior_version(mv.UPLOAD_IMAGE_PARAMS))
    @validation.schema(volume_action.volume_upload_image_v31,
                       mv.UPLOAD_IMAGE_PARAMS)
    def _volume_upload_image(self, req, id, body):
        """Uploads the specified volume to image service."""
        context = req.environ['cinder.context']
        params = body['os-volume_upload_image']
        req_version = req.api_version_request

        force = params.get('force', 'False')
        force = strutils.bool_from_string(force, strict=True)

        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        context.authorize(policy.UPLOAD_IMAGE_POLICY)

        disk_format = params.get("disk_format", "raw")
        image_metadata = {"container_format": params.get(
            "container_format", "bare"),
            "disk_format": disk_format,
            "name": params["image_name"]}

        if volume.encryption_key_id:
            # encrypted volumes cannot be converted on upload
            if (image_metadata['disk_format'] != 'raw'
                    or image_metadata['container_format'] != 'bare'):
                msg = _("An encrypted volume uploaded as an image must use "
                        "'raw' disk_format and 'bare' container_format, "
                        "which are the defaults for these options.")
                raise webob.exc.HTTPBadRequest(explanation=msg)

            # Clone volume encryption key: the current key cannot
            # be reused because it will be deleted when the volume is
            # deleted.
            encryption_key_id = volume_utils.clone_encryption_key(
                context, self._key_manager, volume.encryption_key_id)

            image_metadata['cinder_encryption_key_id'] = encryption_key_id
            image_metadata['cinder_encryption_key_deletion_policy'] = \
                'on_image_deletion'

        if req_version >= mv.get_api_version(
                mv.UPLOAD_IMAGE_PARAMS):

            image_metadata['visibility'] = params.get('visibility', 'private')
            image_metadata['protected'] = strutils.bool_from_string(
                params.get('protected', 'False'), strict=True)

            if image_metadata['visibility'] == 'public':
                context.authorize(policy.UPLOAD_PUBLIC_POLICY)

        try:
            response = self.volume_api.copy_volume_to_image(context,
                                                            volume,
                                                            image_metadata,
                                                            force)
        except exception.InvalidVolume as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)
        except ValueError as error:
            raise webob.exc.HTTPBadRequest(explanation=str(error))
        except messaging.RemoteError as error:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': error.exc_type,
                                                 'err_msg': error.value}
            raise webob.exc.HTTPBadRequest(explanation=msg)
        except Exception as error:
            raise webob.exc.HTTPBadRequest(explanation=str(error))
        return {'os-volume_upload_image': response}

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-extend')
    @validation.schema(volume_action.extend)
    def _extend(self, req, id, body):
        """Extend size of volume."""
        context = req.environ['cinder.context']
        req_version = req.api_version_request
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        size = int(body['os-extend']['new_size'])
        try:
            if (req_version.matches(mv.VOLUME_EXTEND_INUSE) and
                    volume.status in ['in-use']):
                self.volume_api.extend_attached_volume(context, volume, size)
            else:
                self.volume_api.extend(context, volume, size)
        except exception.InvalidVolume as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-update_readonly_flag')
    @validation.schema(volume_action.volume_readonly_update)
    def _volume_readonly_update(self, req, id, body):
        """Update volume readonly flag."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        readonly_flag = body['os-update_readonly_flag']['readonly']

        readonly_flag = strutils.bool_from_string(readonly_flag,
                                                  strict=True)

        self.volume_api.update_readonly_flag(context, volume, readonly_flag)

    @wsgi.response(HTTPStatus.ACCEPTED)
    @wsgi.action('os-retype')
    @validation.schema(volume_action.retype)
    def _retype(self, req, id, body):
        """Change type of existing volume."""
        context = req.environ['cinder.context']
        volume = self.volume_api.get(context, id)
        new_type = body['os-retype']['new_type']
        policy = body['os-retype'].get('migration_policy')

        self.volume_api.retype(context, volume, new_type, policy)

    @wsgi.response(HTTPStatus.OK)
    @wsgi.action('os-set_bootable')
    @validation.schema(volume_action.set_bootable)
    def _set_bootable(self, req, id, body):
        """Update bootable status of a volume."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, id)

        bootable = strutils.bool_from_string(
            body['os-set_bootable']['bootable'], strict=True)

        update_dict = {'bootable': bootable}

        self.volume_api.update(context, volume, update_dict)


class Volume_actions(extensions.ExtensionDescriptor):
    """Enable volume actions."""

    name = "VolumeActions"
    alias = "os-volume-actions"
    updated = "2012-05-31T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = VolumeActionsController()
        extension = extensions.ControllerExtension(self, 'volumes', controller)
        return [extension]



# Copyright (C) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random

from cinder.scheduler import weights


class ChanceWeigher(weights.BaseHostWeigher):
    """Chance Weigher assigns random weights to hosts.

    Used to spread volumes randomly across a list of equally suitable hosts.
    """
    def _weigh_object(self, host_state, weight_properties):
        return random.random()



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import quota_classes as quota_class
from cinder.api import validation
from cinder import db
from cinder import exception
from cinder.policies import quota_class as policy
from cinder import quota


QUOTAS = quota.QUOTAS
GROUP_QUOTAS = quota.GROUP_QUOTAS


class QuotaClassSetsController(wsgi.Controller):

    def _format_quota_set(self, quota_class, quota_set):
        """Convert the quota object to a result dict."""

        quota_set['id'] = str(quota_class)

        return dict(quota_class_set=quota_set)

    def show(self, req, id):
        context = req.environ['cinder.context']
        context.authorize(policy.GET_POLICY)
        try:
            db.sqlalchemy.api.authorize_quota_class_context(context, id)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()
        quota_set = QUOTAS.get_class_quotas(context, id)
        group_quota_set = GROUP_QUOTAS.get_class_quotas(context, id)
        quota_set.update(group_quota_set)

        return self._format_quota_set(id, quota_set)

    @validation.schema(quota_class.update_quota_class)
    def update(self, req, id, body):
        context = req.environ['cinder.context']
        context.authorize(policy.UPDATE_POLICY)
        self.validate_string_length(id, 'quota_class_name',
                                    min_length=1, max_length=255)

        quota_class = id

        for key, value in body['quota_class_set'].items():
            try:
                db.quota_class_update(context, quota_class, key, value)
            except exception.QuotaClassNotFound:
                db.quota_class_create(context, quota_class, key, value)
            except exception.AdminRequired:
                raise webob.exc.HTTPForbidden()

        quota_set = QUOTAS.get_class_quotas(context, quota_class)
        group_quota_set = GROUP_QUOTAS.get_class_quotas(context, quota_class)
        quota_set.update(group_quota_set)

        return {'quota_class_set': quota_set}


class Quota_classes(extensions.ExtensionDescriptor):
    """Quota classes management support."""

    name = "QuotaClasses"
    alias = "os-quota-class-sets"
    updated = "2012-03-12T00:00:00+00:00"

    def get_resources(self):
        resources = []

        res = extensions.ResourceExtension('os-quota-class-sets',
                                           QuotaClassSetsController())
        resources.append(res)

        return resources



#    (c)  Copyright  Kioxia Corporation 2021 All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import json
import os
import ssl

import urllib3

from cinder.volume.drivers.kioxia import entities

urllib3.disable_warnings()
RUN_COMMAND_TRIALS = 20
RUN_COMMAND_SLEEP = 0.5


class ClassBuilder(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if value is not None:
                self.__dict__[key] = value

    def to_json(self):
        return json.dumps(
            self,
            default=lambda o: o.__dict__,
            sort_keys=True,
            indent=4)


class JsonToClass(object):

    def __init__(self, json_object, first=False):
        if isinstance(json_object, list):
            self.records = []
            for list_index in range(len(json_object)):
                list_item = JsonToClass(json_object[list_index])
                self.records.append(list_item)
        else:
            if first:
                self.records = None
            self.build_class(json_object)
        if first:
            if 'status' not in json_object:
                self.status = "Success"
            if 'description' not in json_object:
                self.description = "Success."
        pass

    def __getattr__(self, item):
        return "N/A"

    def to_json(self):
        return json.dumps(
            self,
            default=lambda o: o.__dict__,
            sort_keys=True,
            indent=4)

    def __str__(self):
        return json.dumps(self, default=lambda o: o.__dict__)

    def is_exist(self, item):
        if item in self.__dict__.keys() and self.__dict__[item] is not None:
            return True
        return False

    def build_class(self, json_object):
        json_keys = json_object.keys()
        for key in json_keys:
            if isinstance(json_object[key], list):
                self.__dict__[key] = []
                for i in range(len(json_object[key])):
                    if isinstance(json_object[key][i], dict):
                        sub_object = JsonToClass(json_object[key][i])
                    else:
                        sub_object = json_object[key][i]
                    self.__dict__[key].append(sub_object)
                continue
            if not isinstance(json_object[key], dict):
                self.__dict__[key] = json_object[key]
                continue
            self.__dict__[key] = {}
            sub_object = JsonToClass(json_object[key])
            self.__dict__[key] = sub_object


class ProvisionerVisitor(object):
    #
    # Provisioner Visitor
    #

    def __init__(self, http, command_str):
        self.http = http
        self.command_str = command_str

    @abc.abstractmethod
    def visit(self, url):
        return


class ProvisionerGetVisitor(ProvisionerVisitor):
    #
    # Provisioner Get Visitor
    #

    def visit(self, url, token=None):
        r = self.http.request(
            'GET', url, headers={
                "Authorization": "Bearer " + token})
        return r


class ProvisionerPostVisitor(ProvisionerVisitor):
    #
    # Provisioner Post Visitor
    #

    def __init__(self, http, command_str, json_body):
        ProvisionerVisitor.__init__(self, http, command_str)
        self.json_body = json_body

    def visit(self, url, token=None):
        r = self.http.request(
            'POST',
            url,
            body=self.json_body,
            headers={
                'Content-Type': 'application/json',
                "Authorization": "Bearer " +
                                 token})
        return r


class ProvisionerDeleteVisitor(ProvisionerVisitor):
    #
    # Provisioner Delete Visitor
    #

    def __init__(self, http, command_str):
        ProvisionerVisitor.__init__(self, http, command_str)

    def visit(self, url, token=None):
        r = self.http.request(
            'DELETE',
            url,
            body=None,
            headers={
                'Content-Type': 'application/json',
                "Authorization": "Bearer " +
                                 token})
        return r


class ProvisionerPatchVisitor(ProvisionerVisitor):
    #
    # Provisioner Patch Visitor
    #

    def __init__(self, http, command_str, json_body=None):
        ProvisionerVisitor.__init__(self, http, command_str)
        self.json_body = json_body

    def visit(self, url, token=None):
        r = self.http.request(
            'PATCH',
            url,
            body=self.json_body,
            headers={
                'Content-Type': 'application/json',
                "Authorization": "Bearer " +
                                 token})
        return r


class ProvisionerPutVisitor(ProvisionerVisitor):
    #
    # Provisioner Put Visitor
    #

    def __init__(self, http, command_str, json_body):
        ProvisionerVisitor.__init__(self, http, command_str)
        self.json_body = json_body

    def visit(self, url, token=None):
        r = self.http.request(
            'PUT',
            url,
            body=self.json_body,
            headers={
                'Content-Type': 'application/json',
                "Authorization": "Bearer " +
                                 token})
        return r


class ProvisionerPostDataVisitor(ProvisionerVisitor):
    #
    # Provisioner Post Data Visitor
    #

    def __init__(self, http, command_str, path):
        ProvisionerVisitor.__init__(self, http, command_str)
        self.path = path
        self.timeout = 90

    def visit(self, url, token=None):
        binary_data = open(self.path, 'rb').read()
        disposition = "inline; filename=" + os.path.basename(self.path)

        r = self.http.request(
            'POST',
            url,
            body=binary_data,
            headers={
                'Content-Type': 'application/x-gtar',
                'Content-Disposition': disposition,
                "Authorization": "Bearer " +
                                 token},
            timeout=self.timeout)
        return r


class ProvisionerConnector(object):
    #
    # Provisioner Connector
    #

    def __init__(self, ips, port, visitor):
        self.visitor = visitor
        self.ips = ips
        self.port = port

    def visit_provisioner(self, token=None):
        r = None
        if self.ips:
            num_of_ips = len(self.ips)
            if num_of_ips > 0:
                for i in range(num_of_ips):
                    ip = self.ips[i]
                    url = 'https://' + ip + ':' + \
                          str(self.port) + '/' + self.visitor.command_str
                    try:
                        if token is None:
                            token = "Unknown"
                        r = self.visitor.visit(url, token)
                        if r:
                            if i != 0:
                                KioxiaProvisioner.switch_path(i)
                            return r
                    except BaseException:
                        continue
                return r
            return r
        return r


class KioxiaProvisioner(object):
    #
    # REST client class that interacts with a specific Provisioner
    # :type ips: str array
    # :param ips: Provisioner management IPs
    # :type cert: str
    # :param cert: KumoScale keystore pem file full path
    #

    mgmt_ips = []

    def __init__(self, ips, cert, token, port=8090):
        self.mgmt_ips = ips
        self.port = port
        self.user = None
        self.token = token
        if cert is None:
            cert = '/etc/kioxia/ssdtoolbox.pem'
        KioxiaProvisioner.mgmt_ips = ips
        self.http = urllib3.PoolManager(
            cert_reqs=ssl.CERT_NONE,
            cert_file=cert,
            assert_hostname=False,
            timeout=urllib3.Timeout(
                connect=5.0,
                read=60.0))

    def set_token(self, user, token):
        self.user = user
        self.token = token

    def result_support(self, result):
        if result is not None:
            if result.data is not None:
                if "Status 401" in str(result.data):
                    ClassBuilder()
                    return entities.ProvisionerResponse(
                        None, None, "Bad credentials")
                if "Status 403" in str(result.data):
                    return entities.ProvisionerResponse(
                        None, None, "Access is denied")
                if str(result.data) == "":
                    return entities.ProvisionerResponse([], None, "Success")
                try:
                    result_data = json.loads(result.data)
                    if ('status' in result_data and
                            result_data['status'] != "Success"):
                        return entities.ProvisionerResponse(
                            result_data, None, result_data['status'],
                            result_data['description'])
                    return entities.ProvisionerResponse(result_data)
                except Exception as e:
                    return entities.ProvisionerResponse(
                        None, None, type(e).__name__, e.message)
        return entities.ProvisionerResponse(
            None,
            None,
            "Provisioner Communication Error",
            "Provisioner Communication Error")

    # Call to switch last successful connected ip
    @staticmethod
    def switch_path(ip_idx):
        temp = KioxiaProvisioner.mgmt_ips[0]
        KioxiaProvisioner.mgmt_ips[0] = KioxiaProvisioner.mgmt_ips[ip_idx]
        KioxiaProvisioner.mgmt_ips[ip_idx] = temp

    # Call Provisioner with get request
    def provisioner_get_request(self, api_name):
        get_visitor = ProvisionerGetVisitor(self.http, api_name)
        provisioner_connector = ProvisionerConnector(
            self.mgmt_ips, self.port, get_visitor)
        r = provisioner_connector.visit_provisioner(self.token)
        return self.result_support(r)

    # Call Provisioner with delete request
    def provisioner_delete_request(self, api_name):
        delete_visitor = ProvisionerDeleteVisitor(self.http, api_name)
        provisioner_connector = ProvisionerConnector(
            self.mgmt_ips, self.port, delete_visitor)
        r = provisioner_connector.visit_provisioner(self.token)
        return self.result_support(r)

    # Call Provisioner with patch request
    def provisioner_patch_request(self, api_name, json_body=None):
        patch_visitor = ProvisionerPatchVisitor(self.http, api_name, json_body)
        provisioner_connector = ProvisionerConnector(
            self.mgmt_ips, self.port, patch_visitor)
        r = provisioner_connector.visit_provisioner(self.token)
        return self.result_support(r)

    # Call Provisioner with update request
    def provisioner_put_request(self, api_name, json_body):
        put_visitor = ProvisionerPutVisitor(self.http, api_name, json_body)
        provisioner_connector = ProvisionerConnector(
            self.mgmt_ips, self.port, put_visitor)
        r = provisioner_connector.visit_provisioner(self.token)
        return self.result_support(r)

    # Call Provisioner with post request
    def provisioner_post_request(self, api_name, json_body, password=None):
        post_visitor = ProvisionerPostVisitor(self.http, api_name, json_body)
        provisioner_connector = ProvisionerConnector(
            KioxiaProvisioner.mgmt_ips, self.port, post_visitor)
        r = provisioner_connector.visit_provisioner(self.token)
        return self.result_support(r)

    def get_info(self):
        # Call to Get Info API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain Provisioner information
        #

        result_response = self.provisioner_get_request('info')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity)
        return result_response

    def get_provisioner_info(self):
        # Call to Get Info API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain Provisioner information
        #

        result_response = self.provisioner_get_request('info')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return result_entity
        return result_response

    def add_backend(self, backend_entity):
        # Call to Add Backend API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        j = backend_entity.to_json()
        result_response = self.provisioner_post_request('backends', j)
        return result_response

    def update_backend(self, backend_entity, persistent_id):
        # all to Update Backend API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        j = backend_entity.to_json()
        result_response = self.provisioner_put_request(
            'backends/' + persistent_id, j)
        return result_response

    def delete_backend(self, persistent_id):
        # Call to Delete Backend API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_delete_request(
            'backends/' + persistent_id)
        return result_response

    def get_backends(self):
        # Call to List of Backends API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Backends
        #

        result_response = self.provisioner_get_request('backends')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_backend_by_id(self, uuid):
        # Call to List of Backends API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Backends
        #

        result_response = self.provisioner_get_request('backends/' + uuid)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_volumes(self, tenant_uuid=None):
        # Call to List of Volumes API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_get_request(tenant_id + 'volumes')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_volumes_by_alias(self, alias, tenant_uuid=None):
        # Call to List of Volumes API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_get_request(
            tenant_id + 'volumes_by_alias/' + alias)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_volumes_by_uuid(
            self,
            volume_uuid,
            tenant_uuid=None):
        # Call to List of Volumes API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_get_request(
            tenant_id + 'volumes/' + volume_uuid)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def add_replica(
            self,
            replica_entity,
            volume_uuid,
            tenant_uuid=None):
        # Call to Add Replica API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        j = replica_entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'replica/' + volume_uuid, j)
        return result_response

    def delete_replica(
            self,
            volume_uuid,
            replica_uuid,
            tenant_uuid=None):
        # Call to Delete Replica API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_patch_request(
            tenant_id + 'replica/' + volume_uuid + "/" + replica_uuid)
        return result_response

    def delete_replica_confirm(
            self,
            volume_uuid,
            replica_uuid,
            tenant_uuid=None):
        # Call to Delete Replica Confirm API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_delete_request(
            tenant_id + 'replica/' + volume_uuid + "/" + replica_uuid)
        return result_response

    def create_volume(self, volume_entity, tenant_uuid=None):
        # Call to Create Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        j = volume_entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'volumes', j)
        return result_response

    def delete_volume(self, volume_uuid, tenant_uuid=None):
        # Call to Delete Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_delete_request(
            tenant_id + 'volumes/' + volume_uuid)
        return result_response

    def expand_volume(
            self,
            new_capacity,
            volume_uuid,
            tenant_uuid=None):
        # Call to Expand Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        entity = ClassBuilder(newCapacity=str(new_capacity))
        j = entity.to_json()
        result_response = self.provisioner_patch_request(
            tenant_id + 'volumes/' + volume_uuid, j)
        return result_response

    def set_replica_state(
            self,
            volume_uuid,
            replica_uuid,
            state,
            tenant_uuid=None):
        # Call to Set Replica State API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_patch_request(
            tenant_id + 'replica/' + volume_uuid + "/" +
            replica_uuid + "/" + str(state))
        return result_response

    def get_snapshots(
            self,
            snapshot_uuid=None,
            tenant_uuid=None):
        # Call to List of Snapshots API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        if snapshot_uuid is None:
            result_response = self.provisioner_get_request(
                tenant_id + 'snapshots')
        else:
            result_response = self.provisioner_get_request(
                tenant_id + 'snapshots/' + snapshot_uuid)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_snapshots_by_vol(
            self,
            volume_uuid,
            tenant_uuid=None):
        # Call to Get Snapshot Information via Volume UUID API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_get_request(
            tenant_id + 'snapshots_by_vol/' + volume_uuid)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_snapshots_by_alias(self, alias, tenant_uuid=None):
        # Call to Get Snapshot Information via alias API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_get_request(
            tenant_id + 'snapshots_by_alias/' + alias)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def set_license(self, license_key):
        # Call to Set License API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        entity = ClassBuilder(license=license_key)
        j = entity.to_json()
        result_response = self.provisioner_post_request('license', j)
        return result_response

    def get_license(self):
        # Call to Get License API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_get_request('license')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity)
        return result_response

    def get_inventory(self):
        # Call to Get Inventory API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_get_request('inventory')
        return result_response

    def reset_inventory(self):
        # Call to Reset Inventory API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_delete_request('reset_inventory')
        return result_response

    def get_syslogs(self):
        # Call to Get Syslogs API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_get_request('syslog')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def create_snapshot(
            self,
            snapshot_entity,
            tenant_uuid=None):
        # Call to Create Snapshot API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        j = snapshot_entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'snapshots', j)
        return result_response

    def delete_snapshot(
            self,
            snapshot_uuid,
            tenant_uuid=None):
        # Call to Delete Snapshot API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        result_response = self.provisioner_delete_request(
            tenant_id + 'snapshots/' + snapshot_uuid)
        return result_response

    def create_snapshot_volume(
            self,
            snapshot_volume_entity,
            tenant_uuid=None):
        # Call to Create Snapshot Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        j = snapshot_volume_entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'snapshot_volumes', j)
        return result_response

    def forward_log(self, forward_entity):
        # Call to Forward Log API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        j = forward_entity.to_json()
        result_response = self.provisioner_post_request('forward_log', j)
        return result_response

    def get_hosts(self):
        # Call to Get Hosts API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_get_request('hosts')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def get_hosts_by_name(self, host_name):
        # Call to Get Hosts API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_get_request(
            'hosts?hostName=' + host_name)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def delete_host(self, host_uuid):
        # Call to Delete Host API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        result_response = self.provisioner_delete_request('hosts/' + host_uuid)
        return result_response

    def get_targets(self, host_uuid, volume_uuid):
        # Call to Get Targets API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        if host_uuid is None and volume_uuid is None:
            return entities.ProvisionerResponse(
                None, None, "ParametersError", "All parameters missing")
        if host_uuid is not None:
            request = "?hostId=" + host_uuid
        else:
            request = "?volId=" + volume_uuid
        if host_uuid is not None and volume_uuid is not None:
            request += "&volId=" + volume_uuid
        result_response = self.provisioner_get_request('targets' + request)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def publish(
            self,
            host_uuid,
            volume_uuid,
            tenant_uuid=None):
        # Call to Pablish API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        entity = ClassBuilder(hostId=host_uuid, volId=volume_uuid)
        j = entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'publish', j)
        return result_response

    def unpublish(
            self,
            host_uuid,
            volume_uuid,
            tenant_uuid=None):
        # Call to UnPablish API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        entity = ClassBuilder(hostId=host_uuid, volId=volume_uuid)
        j = entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'unpublish', j)
        return result_response

    def host_probe(self, host_nqn, host_uuid, host_name,
                   client_type, sw_version, duration_in_sec):
        # Call to Host Probe API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        entity = ClassBuilder(
            hostNqn=host_nqn,
            hostId=host_uuid,
            name=host_name,
            clientType=client_type,
            version=sw_version,
            duration=duration_in_sec)
        j = entity.to_json()
        result_response = self.provisioner_post_request('host_probe', j)
        return result_response

    def migrate_volume(
            self,
            volume_uuid,
            replica_uuid,
            tenant_uuid=None):
        # Call to Migrate Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        entity = ClassBuilder(volId=volume_uuid, repId=replica_uuid)
        j = entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'migrate_volume', j)
        return result_response

    def get_tasks(self, task_id=None, host_id=None):
        # Call to Get Tasks API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        if task_id is not None:
            cmd = "tasks?taskId=" + str(task_id)
        elif host_id is not None:
            cmd = "tasks?hostId=" + str(host_id)
        else:
            cmd = "tasks"
        result_response = self.provisioner_get_request(cmd)
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def remove_task(self, task_id, host_id=None):
        # Call to Remove Task API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        cmd = 'tasks?taskId=' + task_id
        if host_id is not None:
            cmd += "&hostId=" + host_id
        result_response = self.provisioner_delete_request(cmd)
        return result_response

    def update_task(self, task_id, host_id, state=None, progress=None,
                    status=None, description=None, tags=None):
        # Call to Update Task API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        entity = ClassBuilder(
            taskId=task_id,
            hostId=host_id,
            state=state,
            progress=progress,
            taskStatus=status,
            statusDescription=description,
            taskConfiguration=tags)
        j = entity.to_json()
        result_response = self.provisioner_put_request('tasks', j)
        return result_response

    def create_tenant(self, tenant_entity):
        # Call to Create Tenant API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        j = tenant_entity.to_json()
        result_response = self.provisioner_post_request('tenants', j)
        return result_response

    def delete_tenant(self, tenant_uuid):
        # Call to Delete Tenant API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        result_response = self.provisioner_delete_request(
            'tenants/' + tenant_uuid)
        return result_response

    def modify_tenant(self, tenant_entity, tenant_uuid):
        # Call to Modify Tenant API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #
        j = tenant_entity.to_json()
        result_response = self.provisioner_put_request(
            'tenants/' + tenant_uuid, j)
        return result_response

    def get_tenants(self):
        # Call to List of Tenants API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data contain List of Volumes
        #

        result_response = self.provisioner_get_request('tenants')
        if result_response.status == "Success":
            result_entity = JsonToClass(result_response.prov_entities, True)
            return entities.ProvisionerResponse(result_entity.records)
        return result_response

    def clone_volume(self, clone_entity, tenant_uuid=None):
        # Call to Clone Volume API

        # @rtype: ProvisionerResponse
        # @returns: Provisioner response data
        #

        tenant_id = ""
        if tenant_uuid is not None:
            tenant_id = tenant_uuid + "/"
        j = clone_entity.to_json()
        result_response = self.provisioner_post_request(
            tenant_id + 'clone_volume', j)
        return result_response

    def get_non_implemented(self, param1=None, param2=None):
        # Call to Get Not Implemented Answer API

        # @rtype: KSResponse
        # @returns: KumoScale response data
        #
        return entities.ProvisionerResponse(None, None, "Not implemented")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.tests.functional import api_samples_test_base as test_base


class QOSSampleJsonTest(test_base.VolumesSampleBase):
    sample_dir = "qos"

    def setUp(self):
        super(QOSSampleJsonTest, self).setUp()
        self.response = self._do_post('qos-specs', 'qos-create-request')

    def test_qos_create(self):
        self._verify_response('qos-create-response', {}, self.response, 200)

    def test_qos_list(self):
        response = self._do_get('qos-specs')
        self._verify_response('qos-list-response', {}, response, 200)

    def test_qos_show(self):
        res = jsonutils.loads(self.response.content)['qos_specs']
        response = self._do_get('qos-specs/%s' % res['id'])
        self._verify_response('qos-show-response', {}, response, 200)

    def test_qos_update(self):
        res = jsonutils.loads(self.response.content)['qos_specs']
        response = self._do_put('qos-specs/%s' % res['id'],
                                'qos-update-request')
        self._verify_response('qos-update-response', {}, response, 200)

    def test_qos_show_associations(self):
        res = jsonutils.loads(self.response.content)['qos_specs']
        response = self._do_get('qos-specs/%s/associations' % res['id'])
        self._verify_response('qos_show_response', {}, response, 200)
